# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime
from hashlib import sha256
from os import urandom
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union

import sqlalchemy
from alembic import command, op
from alembic.config import Config
from dogpile.cache.api import NoValue
from sqlalchemy import Column, PrimaryKeyConstraint, func, inspect
from sqlalchemy.dialects.postgresql.base import PGInspector
from sqlalchemy.exc import DatabaseError, IntegrityError
from sqlalchemy.orm import declarative_base
from sqlalchemy.schema import CreateSchema, CreateTable, DropConstraint, DropTable, ForeignKeyConstraint, MetaData, Table
from sqlalchemy.sql.ddl import DropSchema
from sqlalchemy.sql.expression import select, text

from rucio import alembicrevision
from rucio.common.cache import MemcacheRegion
from rucio.common.config import config_get, config_get_list
from rucio.common.schema import get_schema_value
from rucio.common.types import InternalAccount, LoggerFunction
from rucio.common.utils import generate_uuid
from rucio.db.sqla import models
from rucio.db.sqla.constants import AccountStatus, AccountType, IdentityType
from rucio.db.sqla.session import get_dump_engine, get_engine, get_session
from rucio.db.sqla.types import InternalScopeString, String

if TYPE_CHECKING:
    from collections.abc import Sequence

    from sqlalchemy.engine import Inspector
    from sqlalchemy.orm import Query, Session

    # TypeVar representing the DeclarativeObj class defined inside _create_temp_table
    DeclarativeObj = TypeVar('DeclarativeObj')

REGION = MemcacheRegion(expiration_time=600, memcached_expire_time=3660)


def build_database() -> None:
    """ Applies the schema to the database. Run this command once to build the database. """
    engine = get_engine()

    schema = config_get('database', 'schema', raise_exception=False, check_config_table=False)
    if schema:
        print('Schema set in config, trying to create schema:', schema)
        try:
            with engine.connect() as conn:
                with conn.begin():
                    conn.execute(CreateSchema(schema))
        except Exception as e:
            print('Cannot create schema, please validate manually if schema creation is needed, continuing:', e)

    models.register_models(engine)

    # Put the database under version control
    alembic_cfg = Config(config_get('alembic', 'cfg'))
    command.stamp(alembic_cfg, "head")


def dump_schema() -> None:
    """ Creates a schema dump to a specific database. """
    engine = get_dump_engine()
    models.register_models(engine)


def destroy_database() -> None:
    """ Removes the schema from the database. Only useful for test cases or malicious intents. """
    engine = get_engine()

    try:
        models.unregister_models(engine)
    except Exception as e:
        print('Cannot destroy schema -- assuming already gone, continuing:', e)


def drop_everything() -> None:
    """
    Pre-gather all named constraints and table names, and drop everything.
    This is better than using metadata.reflect(); metadata.drop_all()
    as it handles cyclical constraints between tables.
    Ref. https://github.com/sqlalchemy/sqlalchemy/wiki/DropEverything
    """
    engine = get_engine()

    # the transaction only applies if the DB supports
    # transactional DDL, i.e. Postgresql, MS SQL Server

    with engine.connect() as conn:

        inspector: Union["Inspector", PGInspector] = inspect(conn)

        for tname, fkcs in reversed(
                inspector.get_sorted_table_and_fkc_names(schema='*')):
            if tname:
                drop_table_stmt = DropTable(Table(tname, MetaData(), schema='*'))
                conn.execute(drop_table_stmt)
            elif fkcs:
                if not engine.dialect.supports_alter:
                    continue
                for tname, fkc in fkcs:
                    fk_constraint = ForeignKeyConstraint((), (), name=fkc)
                    Table(tname, MetaData(), fk_constraint)
                    drop_constraint_stmt = DropConstraint(fk_constraint)
                    conn.execute(drop_constraint_stmt)

        schema = config_get('database', 'schema', raise_exception=False)
        if schema:
            conn.execute(DropSchema(schema, cascade=True))

        if engine.dialect.name == 'postgresql':
            if not isinstance(inspector, PGInspector):
                raise ValueError('expected a PGInspector')
            for enum in inspector.get_enums(schema='*'):
                sqlalchemy.Enum(**enum).drop(bind=conn)


def create_base_vo() -> None:
    """ Creates the base VO """

    session_scoped = get_session()

    vo = models.VO(vo='def', description='Default base VO', email='N/A')
    with session_scoped() as s:
        with s.begin():
            s.add_all([vo])


def create_root_account() -> None:
    """
    Inserts the default root account to an existing database. Make sure to change the default password later.
    """

    multi_vo = bool(config_get('common', 'multi_vo', False, False))

    up_id = config_get('bootstrap', 'userpass_identity', default='ddmlab')
    up_pwd = config_get('bootstrap', 'userpass_pwd', default='secret')
    up_email = config_get('bootstrap', 'userpass_email', default='ph-adp-ddm-lab@cern.ch')
    x509_id = config_get('bootstrap', 'x509_identity', default='emailAddress=ph-adp-ddm-lab@cern.ch,CN=DDMLAB Client Certificate,OU=PH-ADP-CO,O=CERN,ST=Geneva,C=CH')
    x509_email = config_get('bootstrap', 'x509_email', default='ph-adp-ddm-lab@cern.ch')
    gss_id = config_get('bootstrap', 'gss_identity', default='ddmlab@CERN.CH')
    gss_email = config_get('bootstrap', 'gss_email', default='ph-adp-ddm-lab@cern.ch')
    ssh_id = config_get('bootstrap', 'ssh_identity',
                        default='ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAQEAq5LySllrQFpPL614sulXQ7wnIr1aGhGtl8b+HCB/'
                        '0FhMSMTHwSjX78UbfqEorZV16rXrWPgUpvcbp2hqctw6eCbxwqcgu3uGWaeS5A0iWRw7oXUh6ydn'
                        'Vy89zGzX1FJFFDZ+AgiZ3ytp55tg1bjqqhK1OSC0pJxdNe878TRVVo5MLI0S/rZY2UovCSGFaQG2'
                        'iLj14wz/YqI7NFMUuJFR4e6xmNsOP7fCZ4bGMsmnhR0GmY0dWYTupNiP5WdYXAfKExlnvFLTlDI5'
                        'Mgh4Z11NraQ8pv4YE1woolYpqOc/IMMBBXFniTT4tC7cgikxWb9ZmFe+r4t6yCDpX4IL8L5GOQ== ddmlab'
                        )
    ssh_email = config_get('bootstrap', 'ssh_email', default='ph-adp-ddm-lab@cern.ch')

    session_scoped = get_session()

    if multi_vo:
        access = 'super_root'
    else:
        access = 'root'

    account = models.Account(account=InternalAccount(access, 'def'), account_type=AccountType.SERVICE, status=AccountStatus.ACTIVE)

    salt = urandom(255)
    salted_password = salt + up_pwd.encode()
    hashed_password = sha256(salted_password).hexdigest()

    identities = []
    associations = []

    if up_id and up_pwd:
        identity1 = models.Identity(identity=up_id, identity_type=IdentityType.USERPASS, password=hashed_password, salt=salt, email=up_email)
        iaa1 = models.IdentityAccountAssociation(identity=identity1.identity, identity_type=identity1.identity_type, account=account.account, is_default=True)
        identities.append(identity1)
        associations.append(iaa1)

    # X509 authentication
    if x509_id and x509_email:
        identity2 = models.Identity(identity=x509_id, identity_type=IdentityType.X509, email=x509_email)
        iaa2 = models.IdentityAccountAssociation(identity=identity2.identity, identity_type=identity2.identity_type, account=account.account, is_default=True)
        identities.append(identity2)
        associations.append(iaa2)

    # GSS authentication
    if gss_id and gss_email:
        identity3 = models.Identity(identity=gss_id, identity_type=IdentityType.GSS, email=gss_email)
        iaa3 = models.IdentityAccountAssociation(identity=identity3.identity, identity_type=identity3.identity_type, account=account.account, is_default=True)
        identities.append(identity3)
        associations.append(iaa3)

    # SSH authentication
    if ssh_id and ssh_email:
        identity4 = models.Identity(identity=ssh_id, identity_type=IdentityType.SSH, email=ssh_email)
        iaa4 = models.IdentityAccountAssociation(identity=identity4.identity, identity_type=identity4.identity_type, account=account.account, is_default=True)
        identities.append(identity4)
        associations.append(iaa4)

    with session_scoped() as s:
        s.begin()
        # Apply
        for identity in identities:
            try:
                s.add(identity)
                s.commit()
            except IntegrityError:
                # Identities may already be in the DB when running multi-VO conversion
                s.rollback()
        s.add(account)
        s.flush()
        s.add_all(associations)
        s.commit()


def get_db_time() -> Optional[datetime]:
    """ Gives the utc time on the db. """
    session_scoped = get_session()
    try:
        storage_date_format = None
        if session_scoped.bind.dialect.name == 'oracle':
            query = select(text("sys_extract_utc(systimestamp)"))
        elif session_scoped.bind.dialect.name == 'mysql':
            query = select(text("utc_timestamp()"))
        elif session_scoped.bind.dialect.name == 'sqlite':
            query = select(text("datetime('now', 'utc')"))
            storage_date_format = '%Y-%m-%d  %H:%M:%S'
        else:
            query = select(func.current_date())

        session = session_scoped()
        session.begin()
        for now, in session.execute(query):
            if storage_date_format:
                return datetime.strptime(now, storage_date_format)
            return now
    finally:
        session_scoped.remove()


def get_count(q: "Query") -> int:
    """
    Fast way to get count in SQLAlchemy
    Source: https://gist.github.com/hest/8798884
    Some limits, see a more thorough version above
    """

    count_q = q.statement.with_only_columns([func.count()]).order_by(None)
    count = q.session.execute(count_q).scalar()
    return count


def is_old_db() -> bool:
    """
    Returns true, if alembic is used and the database is not on the
    same revision as the code base.
    """
    schema = config_get('database', 'schema', raise_exception=False)

    # checks if alembic is being used by looking up the AlembicVersion table
    inspector = inspect(get_engine())
    if not inspector.has_table(models.AlembicVersion.__tablename__, schema):
        return False

    session_scoped = get_session()
    with session_scoped() as s:
        with s.begin():
            # query = s.query(models.AlembicVersion.version_num)
            query = s.execute(select(models.AlembicVersion)).scalars().all()
            # return query.count() != 0 and str(query.first()[0]) != alembicrevision.ALEMBIC_REVISION
            return (len(query) != 0 and str(query[0].version_num) != alembicrevision.ALEMBIC_REVISION)


def json_implemented(*, session: Optional["Session"] = None) -> bool:
    """
    Checks if the database on the current server installation can support json fields.

    :param session: The active session of the database.
    :type session: Optional[Session]
    :returns: True, if json is supported, False otherwise.
    """
    if session is None:
        session = get_session()

    if session.bind.dialect.name == 'oracle':
        oracle_version = int(session.connection().connection.version.split('.')[0])
        if oracle_version < 12:
            return False
    elif session.bind.dialect.name == 'sqlite':
        return False

    return True


def try_drop_constraint(constraint_name: str, table_name: str) -> None:
    """
    Tries to drop the given constrained and returns successfully if the
    constraint already existed on Oracle databases.

    :param constraint_name: the constraint's name
    :param table_name: the table name where the constraint resides
    """
    try:
        op.drop_constraint(constraint_name, table_name)
    except DatabaseError as e:
        if 'nonexistent constraint' not in str(e):
            raise RuntimeError(e)


def list_oracle_global_temp_tables(session: "Session") -> list[str]:
    """
    Retrieve the list of global temporary tables in oracle
    """
    global_temp_tables = config_get_list('core', 'oracle_global_temp_tables', raise_exception=False, check_config_table=False, default='')
    if global_temp_tables:
        return [t.upper() for t in global_temp_tables]

    cache_key = 'oracle_global_temp_tables'
    # Set long expiration time to avoid hammering the database with this costly query
    global_temp_tables = REGION.get(cache_key, expiration_time=3600)
    if isinstance(global_temp_tables, NoValue):
        # As of time of writing, get_temp_table_names doesn't allow setting the correct schema when called
        # (like get_table_names allows). This may be fixed in a later version of sqlalchemy:
        # FIXME: substitute with something like this:
        # global_temp_tables = [t.upper() for t in inspect(session.bind).get_temp_table_names()]
        global_temp_tables = [
            str(t[0]).upper()
            for t in session.execute(
                text('SELECT UPPER(table_name) '
                     'FROM all_tables '
                     'WHERE OWNER = :owner AND IOT_NAME IS NULL AND DURATION IS NOT NULL'),
                dict(owner=models.BASE.metadata.schema.upper())
            )
        ]
        REGION.set(cache_key, global_temp_tables)
    return global_temp_tables


def _create_temp_table(
        name: str,
        *columns: "Sequence[Column]",
        primary_key: Optional["Sequence[Any]"] = None,
        oracle_global_name: Optional[str] = None,
        session: Optional["Session"] = None,
        logger: LoggerFunction = logging.log
) -> type["DeclarativeObj"]:
    """
    Create a temporary table with the given columns, register it into a declarative base, and return it.

    Attention! calling this function while a table with the same `name` is being used will lead to
    bad consequences. Don't use it in recursive calls without taking special care.

    Declarative definition _requires_ a primary key. It should be a subset of '*columns' argument
    (either a single column, or a list). If not explicitly give, will use the first column as primary key.

    On oracle, there are 2 possible types of temporary tables: global/private.
    In the global case, tables are created once and then can be used by any session (with private data).
    Private tables are created on-the fly, but have many limitations. For example: no indexes allowed.
    This primary key is "fake" in this case, because it only exists in sqlalchemy and not in the database.

    Mysql and sqlite don't support automatic cleanup of temporary tables on commit. This means that a
    temporary table definition is preserved for the lifetime of a session. A session is regularly
    re-used by sqlalchemy, that's why we have to assume the required temporary table already exist and
    could contain data from a previous transaction. Drop all data from that table.
    """
    if not primary_key:
        primary_key = columns[0]
    if not hasattr(primary_key, '__iter__'):
        primary_key = (primary_key, )

    oracle_table_is_global = False
    if session.bind.dialect.name == 'oracle':
        # Retrieve the list of global temporary tables on oracle.
        # If the requested table is found to be global, reuse it,
        # otherwise create a private temporary table with random name
        global_temp_tables = list_oracle_global_temp_tables(session=session)
        if oracle_global_name is None:
            oracle_global_name = name
        if oracle_global_name.upper() in global_temp_tables:
            oracle_table_is_global = True
            additional_kwargs = {
                'oracle_on_commit': 'DELETE ROWS',
                'prefixes': ['GLOBAL TEMPORARY'],
            }
        else:
            logger(logging.WARNING, f"Global temporary table {name} doesn't exist. Using private temporary table.")
            additional_kwargs = {
                'oracle_on_commit': 'DROP DEFINITION',
                'prefixes': ['PRIVATE TEMPORARY'],
            }
            # PRIVATE_TEMP_TABLE_PREFIX, which defaults to "ORA$PTT_", _must_ prefix the name
            name = f"ORA$PTT_{name}"
            # Oracle doesn't support the if_not_exists construct, so add a random suffix to the
            # name to allow multiple calls to the same function within the same session.
            # For example: multiple attach_dids_to_dids(..., session=session)
            name = f'{name}_{generate_uuid()}'
    elif session.bind.dialect.name == 'postgresql':
        additional_kwargs = {
            'postgresql_on_commit': 'DROP',
            'prefixes': ['TEMPORARY'],
        }
    else:
        additional_kwargs = {
            'prefixes': ['TEMPORARY'],
        }

    base = declarative_base()
    table = Table(
        oracle_global_name if oracle_table_is_global else name,
        base.metadata,
        *columns,
        schema=models.BASE.metadata.schema if oracle_table_is_global else None,  # Temporary tables exist in a special schema, so a schema name cannot be given when creating a temporary table
        **additional_kwargs,
    )

    # Oracle private temporary tables don't support indexes.
    # So skip adding the constraints to the table in that case.
    if not session.bind.dialect.name == 'oracle' or oracle_table_is_global:
        table.append_constraint(PrimaryKeyConstraint(*primary_key))

    class DeclarativeObj(base):
        __table__ = table
        # The declarative base requires a primary key, even if it doesn't exist in the database.
        __mapper_args__ = {
            "primary_key": primary_key,
        }

    # Ensure the table exists and is empty.
    if session.bind.dialect.name == 'oracle':
        # Oracle doesn't support if_not_exists.
        # We ensured the unicity by appending a random string to the table name.
        if not oracle_table_is_global:
            session.execute(CreateTable(table))
    elif session.bind.dialect.name == 'postgresql':
        session.execute(CreateTable(table))
    else:
        # If it already exists, it can contain leftover data from a previous transaction
        # executed by sqlalchemy within the same session (which is being re-used now)
        # This is not the case for oracle and postgresql thanks to their "on_commit" support.
        session.execute(CreateTable(table, if_not_exists=True))
        session.query(DeclarativeObj).delete()
    return DeclarativeObj


class TempTableManager:
    """
    A class which manages temporary tables created during a session.

    Attempts to create multiple temporary tables with the same name during a session will
    result in creation of unique tables with an integer "index" suffix added to their name.
    Without this, there would be a risk that a temporary table containing needed data are
    cleaned up during a recursive function call, resulting in unexpected behavior.
    The recursive call may be indirect and hard to catch. For example:
    functionA -> functionB -> functionC -> functionA

    The lifecycle of this object is bound to a particular session. In rucio, we naver use
    sessions in multiple threads at a time, so no need to protect indexes with a mutex.
    """

    def __init__(self, session: "Session"):
        self.session = session

        self.next_idx_to_use = {}

    def create_temp_table(
            self,
            name: str,
            *columns: "Sequence[Column]",
            primary_key: Optional["Sequence[Any]"] = None,
            logger: LoggerFunction = logging.log
    ) -> type["DeclarativeObj"]:
        idx = self.next_idx_to_use.setdefault(name, 0)
        table = _create_temp_table(f'{name}_{idx}', *columns, primary_key=primary_key, session=self.session, logger=logger)
        self.next_idx_to_use[name] = idx + 1
        return table

    def create_scope_name_table(self, logger: LoggerFunction = logging.log) -> type["DeclarativeObj"]:
        """
        Create a temporary table with columns 'scope' and 'name'
        """

        columns = [
            Column("scope", InternalScopeString(get_schema_value('SCOPE_LENGTH'))),
            Column("name", String(get_schema_value('NAME_LENGTH'))),
        ]
        return self.create_temp_table(
            'TEMPORARY_SCOPE_NAME',
            *columns,
            primary_key=columns,
            logger=logger,
        )

    def create_association_table(self, logger: LoggerFunction = logging.log) -> type["DeclarativeObj"]:
        """
        Create a temporary table with columns 'scope', 'name', 'child_scope'and 'child_name'
        """

        columns = [
            Column("scope", InternalScopeString(get_schema_value('SCOPE_LENGTH'))),
            Column("name", String(get_schema_value('NAME_LENGTH'))),
            Column("child_scope", InternalScopeString(get_schema_value('SCOPE_LENGTH'))),
            Column("child_name", String(get_schema_value('NAME_LENGTH'))),
        ]
        return self.create_temp_table(
            'TEMPORARY_ASSOCIATION',
            *columns,
            primary_key=columns,
            logger=logger,
        )

    def create_id_table(self, logger: LoggerFunction = logging.log) -> type["DeclarativeObj"]:
        """
        Create a temp table with a single id column of uuid type
        """

        return self.create_temp_table(
            'TEMPORARY_ID',
            Column("id", models.GUID()),
            logger=logger,
        )


def temp_table_mngr(session: "Session") -> TempTableManager:
    """
    Creates (if doesn't yet exist) and returns a TempTableManager instance associated to the session
    """
    key = 'temp_table_mngr'
    mngr = session.info.get(key)
    if not mngr:
        mngr = TempTableManager(session)
        session.info[key] = mngr
    return mngr



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Union

import rucio.common.exception
import rucio.gateway.permission
from rucio.common.types import InternalAccount, RSEResolvedGlobalAccountLimitDict
from rucio.common.utils import gateway_update_return_dict
from rucio.core import account_limit as account_limit_core
from rucio.core.account import account_exists
from rucio.core.rse import get_rse_id, get_rse_name
from rucio.db.sqla.session import read_session, transactional_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


@read_session
def get_rse_account_usage(
    rse: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> list[dict[str, Any]]:
    """
    Returns the account limit and usage for all for all accounts on a RSE.

    :param rse:      The RSE name.
    :param vo:       The VO to act on.
    :param session:  The database session in use.
    :return:         List of dictionaries.
    """
    rse_id = get_rse_id(rse=rse, vo=vo, session=session)

    return [gateway_update_return_dict(d, session=session) for d in account_limit_core.get_rse_account_usage(rse_id=rse_id, session=session)]


@read_session
def get_local_account_limits(
    account: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> dict[str, Any]:
    """
    Lists the limitation names/values for the specified account name.

    REST API: http://<host>:<port>/rucio/account/<account>/limits

    :param account:     The account name.
    :param vo:          The VO to act on.
    :param session:     The database session in use.

    :returns: The account limits.
    """

    internal_account = InternalAccount(account, vo=vo)

    rse_instead_id = {}
    for elem in account_limit_core.get_local_account_limits(account=internal_account, session=session).items():
        rse_instead_id[get_rse_name(rse_id=elem[0], session=session)] = elem[1]
    return rse_instead_id


@read_session
def get_local_account_limit(
    account: str,
    rse: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> dict[str, Union[int, float, None]]:
    """
    Lists the limitation names/values for the specified account name and rse name.

    REST API: http://<host>:<port>/rucio/account/<account>/limits

    :param account:     The account name.
    :param rse:         The rse name.
    :param vo:          The VO to act on.
    :param session:     The database session in use.

    :returns: The account limit.
    """

    internal_account = InternalAccount(account, vo=vo)

    rse_id = get_rse_id(rse=rse, vo=vo, session=session)
    return {rse: account_limit_core.get_local_account_limit(account=internal_account, rse_id=rse_id, session=session)}


@read_session
def get_global_account_limits(
    account: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> dict[str, RSEResolvedGlobalAccountLimitDict]:
    """
    Lists the limitation names/values for the specified account name.

    REST API: http://<host>:<port>/rucio/account/<account>/limits

    :param account:     The account name.
    :param vo:          The VO to act on.
    :param session:     The database session in use.

    :returns: The account limits.
    """
    if account:
        internal_account = InternalAccount(account, vo=vo)
    else:
        internal_account = InternalAccount('*', vo=vo)

    return account_limit_core.get_global_account_limits(account=internal_account, session=session)


@read_session
def get_global_account_limit(
    account: str,
    rse_expression: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> dict[str, dict[str, RSEResolvedGlobalAccountLimitDict]]:
    """
    Lists the limitation names/values for the specified account name and rse expression.

    REST API: http://<host>:<port>/rucio/account/<account>/limits

    :param account:         The account name.
    :param rse_expression:  The rse expression.
    :param vo:              The VO to act on.
    :param session:         The database session in use.

    :returns: The account limit.
    """

    internal_account = InternalAccount(account, vo=vo)

    return {rse_expression: account_limit_core.get_global_account_limit(account=internal_account, rse_expression=rse_expression, session=session)}


@transactional_session
def set_local_account_limit(
    account: str,
    rse: str,
    bytes_: int,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Set an account limit.

    :param account: The account name.
    :param rse:     The rse name.
    :param bytes_:   The limit in bytes.
    :param issuer:  The issuer account_core.
    :param vo:      The VO to act on.
    :param session: The database session in use.
    """
    rse_id = get_rse_id(rse=rse, vo=vo, session=session)

    kwargs = {'account': account, 'rse': rse, 'rse_id': rse_id, 'bytes': bytes_}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='set_local_account_limit', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise rucio.common.exception.AccessDenied('Account %s can not set account limits. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)

    if not account_exists(account=internal_account, session=session):
        raise rucio.common.exception.AccountNotFound('Account %s does not exist' % (internal_account))

    account_limit_core.set_local_account_limit(account=internal_account, rse_id=rse_id, bytes_=bytes_, session=session)


@transactional_session
def set_global_account_limit(
    account: str,
    rse_expression: str,
    bytes_: int,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> None:
    """
    Set a global account limit.

    :param account:         The account name.
    :param rse_expression:  The rse expression.
    :param bytes_:           The limit in bytes.
    :param issuer:          The issuer account_core.
    :param vo:              The VO to act on.
    :param session:         The database session in use.
    """

    kwargs = {'account': account, 'rse_expression': rse_expression, 'bytes': bytes_}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='set_global_account_limit', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise rucio.common.exception.AccessDenied('Account %s can not set account limits. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)

    if not account_exists(account=internal_account, session=session):
        raise rucio.common.exception.AccountNotFound('Account %s does not exist' % (internal_account))

    account_limit_core.set_global_account_limit(account=internal_account, rse_expression=rse_expression, bytes_=bytes_, session=session)


@transactional_session
def delete_local_account_limit(
    account: str,
    rse: str,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> bool:
    """
    Delete an account limit.

    :param account: The account name.
    :param rse:     The rse name.
    :param issuer:  The issuer account_core.
    :param vo:      The VO to act on.
    :param session: The database session in use.

    :returns: True if successful; False otherwise.
    """

    rse_id = get_rse_id(rse=rse, vo=vo, session=session)
    kwargs = {'account': account, 'rse': rse, 'rse_id': rse_id}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='delete_local_account_limit', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise rucio.common.exception.AccessDenied('Account %s can not delete account limits. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)

    if not account_exists(account=internal_account, session=session):
        raise rucio.common.exception.AccountNotFound('Account %s does not exist' % (internal_account))

    return account_limit_core.delete_local_account_limit(account=internal_account, rse_id=rse_id, session=session)


@transactional_session
def delete_global_account_limit(
    account: str,
    rse_expression: str,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> bool:
    """
    Delete a global account limit..

    :param account:        The account name.
    :param rse_expression: The rse expression.
    :param issuer:         The issuer account_core.
    :param vo:             The VO to act on.
    :param session:        The database session in use.

    :returns: True if successful; False otherwise.
    """

    kwargs = {'account': account, 'rse_expression': rse_expression}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='delete_global_account_limit', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise rucio.common.exception.AccessDenied('Account %s can not delete global account limits. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)

    if not account_exists(account=internal_account, session=session):
        raise rucio.common.exception.AccountNotFound('Account %s does not exist' % (internal_account))

    return account_limit_core.delete_global_account_limit(account=internal_account, rse_expression=rse_expression, session=session)


@read_session
def get_local_account_usage(
    account: str,
    rse: str,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> list[dict[str, Any]]:
    """
    Get the account usage and connect it with (if available) the account limits of the account.

    :param account:  The account to read.
    :param rse:      The rse to read (If none, get all).
    :param issuer:   The issuer account.
    :param vo:       The VO to act on.
    :param session:  The database session in use.

    :returns:        List of dicts {'rse_id', 'rse', 'bytes', 'files', 'bytes_limit', 'bytes_remaining'}
    """

    rse_id = None

    if rse:
        rse_id = get_rse_id(rse=rse, vo=vo, session=session)
    kwargs = {'account': account, 'rse': rse, 'rse_id': rse_id}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='get_local_account_usage', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise rucio.common.exception.AccessDenied('Account %s can not list account usage. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)

    if not account_exists(account=internal_account, session=session):
        raise rucio.common.exception.AccountNotFound('Account %s does not exist' % (internal_account))

    return [gateway_update_return_dict(d, session=session) for d in account_limit_core.get_local_account_usage(account=internal_account, rse_id=rse_id, session=session)]


@read_session
def get_global_account_usage(
    account: str,
    rse_expression: str,
    issuer: str,
    vo: str = 'def',
    *,
    session: "Session"
) -> list[dict[str, Any]]:
    """
    Get the account usage and connect it with (if available) the account limits of the account.

    :param account:         The account to read.
    :param rse_expression:  The rse expression to read (If none, get all).
    :param issuer:          The issuer account.
    :param vo:              The VO to act on.
    :param session:         The database session in use.

    :returns:               List of dicts {'rse_id', 'rse', 'bytes', 'files', 'bytes_limit', 'bytes_remaining'}
    """

    kwargs = {'account': account, 'rse_expression': rse_expression}
    auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='get_global_account_usage', kwargs=kwargs, session=session)
    if not auth_result.allowed:
        raise rucio.common.exception.AccessDenied('Account %s can not list global account usage. %s' % (issuer, auth_result.message))

    internal_account = InternalAccount(account, vo=vo)

    if not account_exists(account=internal_account, session=session):
        raise rucio.common.exception.AccountNotFound('Account %s does not exist' % (internal_account))

    return [gateway_update_return_dict(d, session=session) for d in account_limit_core.get_global_account_usage(account=internal_account, rse_expression=rse_expression, session=session)]



#!/usr/bin/env python3
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import sys

base_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(base_path)
os.chdir(base_path)

from argparse import ArgumentParser  # noqa: E402

from rucio.db.sqla.util import build_database, create_base_vo, create_root_account, destroy_database, drop_everything  # noqa: E402

if __name__ == '__main__':

    parser = ArgumentParser()
    parser.add_argument('-d', '--drop-everything', action="store_true", default=False, help='Drop all tables and constraints')
    args = parser.parse_args()

    if args.drop_everything:
        drop_everything()
    else:
        destroy_database()

    build_database()
    create_base_vo()
    create_root_account()



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Collector to get the SRM free and used information for DATADISK RSEs.
"""

from typing import TYPE_CHECKING, Optional

from sqlalchemy import and_, select

from rucio.db.sqla.models import RSEAttrAssociation, RSEUsage
from rucio.db.sqla.session import read_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


class FreeSpaceCollector:
    """
    Collector to get the SRM free and used information for DATADISK RSEs.
    """
    class _FreeSpaceCollector:
        """
        Hidden implementation
        """
        def __init__(self):
            self.rses = {}

        @read_session
        def _collect_free_space(
            self,
            *,
            session: Optional["Session"] = None
        ) -> None:
            """
            Retrieve free space from database
            """
            stmt = select(
                RSEUsage.rse_id,
                RSEUsage.free,
                RSEUsage.used
            ).join(
                RSEAttrAssociation,
                RSEAttrAssociation.rse_id == RSEUsage.rse_id
            ).where(
                and_(RSEUsage.source == 'storage',
                     RSEAttrAssociation.key == 'type',
                     RSEAttrAssociation.value == 'DATADISK'),
            )
            for rse_id, free, used in session.execute(stmt).all():  # type: ignore (session could be None)
                self.rses[rse_id] = {'total': used + free, 'used': used, 'free': free}

    instance = None

    def __init__(self):
        if not FreeSpaceCollector.instance:
            FreeSpaceCollector.instance = FreeSpaceCollector._FreeSpaceCollector()

    def collect_free_space(self) -> None:
        """
        Execute the free space collector
        """
        self.instance._collect_free_space()  # type: ignore

    def get_rse_space(self) -> dict[str, dict[str, int]]:
        """
        Return the RSE space
        """
        return self.instance.rses  # type: ignore



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from rucio.common import exception
from rucio.common.checksum import PREFERRED_CHECKSUM
from rucio.common.utils import execute
from rucio.rse.protocols import protocol


class Default(protocol.RSEProtocol):
    """ Implementing access to RSEs using the XRootD protocol using GSI authentication."""

    @property
    def _auth_env(self):
        if self.auth_token:
            return f"XrdSecPROTOCOL=ztn BEARER_TOKEN='{self.auth_token}'"
        else:
            return 'XrdSecPROTOCOL=gsi'

    def __init__(self, protocol_attr, rse_settings, logger=logging.log):
        """ Initializes the object with information about the referred RSE.

            :param props: Properties derived from the RSE Repository
        """
        super(Default, self).__init__(protocol_attr, rse_settings, logger=logger)

        self.scheme = self.attributes['scheme']
        self.hostname = self.attributes['hostname']
        self.port = str(self.attributes['port'])
        self.logger = logger

    def path2pfn(self, path):
        """
            Returns a fully qualified PFN for the file referred by path.

            :param path: The path to the file.

            :returns: Fully qualified PFN.

        """
        self.logger(logging.DEBUG, 'xrootd.path2pfn: path: {}'.format(path))
        if not path.startswith('xroot') and not path.startswith('root'):
            if path.startswith('/'):
                return '%s://%s:%s/%s' % (self.scheme, self.hostname, self.port, path)
            else:
                return '%s://%s:%s//%s' % (self.scheme, self.hostname, self.port, path)
        else:
            return path

    def exists(self, pfn):
        """ Checks if the requested file is known by the referred RSE.

            :param pfn: Physical file name

            :returns: True if the file exists, False if it doesn't

            :raise  ServiceUnavailable
        """
        self.logger(logging.DEBUG, 'xrootd.exists: pfn: {}'.format(pfn))
        try:
            path = self.pfn2path(pfn)
            cmd = f'{self._auth_env} xrdfs {self.hostname}:{self.port} stat {path}'
            self.logger(logging.DEBUG, 'xrootd.exists: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                return False
        except Exception as e:
            raise exception.ServiceUnavailable(e)

        return True

    def stat(self, path):
        """
        Returns the stats of a file.

        :param path: path to file

        :raises ServiceUnavailable: if some generic error occurred in the library.

        :returns: a dict with two keys, filesize and an element of GLOBALLY_SUPPORTED_CHECKSUMS.
        """
        self.logger(logging.DEBUG, f'xrootd.stat: path: {path}')
        ret = {}
        chsum = None
        if path.startswith('root:'):
            path = self.pfn2path(path)

        try:
            # xrdfs stat for getting filesize
            cmd = f'{self._auth_env} xrdfs {self.hostname}:{self.port} stat {path}'
            self.logger(logging.DEBUG, 'xrootd.stat: filesize cmd: {}'.format(cmd))
            status_stat, out, err = execute(cmd)
            if status_stat == 0:
                for line in out.split('\n'):
                    if line and ':' in line:
                        k, v = line.split(':', maxsplit=1)
                        if k.strip().lower() == 'size':
                            ret['filesize'] = v.strip()
                            break

            # xrdfs query checksum for getting checksum
            cmd = f'{self._auth_env} xrdfs {self.hostname}:{self.port} query checksum {path}'
            self.logger(logging.DEBUG, 'xrootd.stat: checksum cmd: {}'.format(cmd))
            status_query, out, err = execute(cmd)
            if status_query == 0:
                chsum, value = out.strip('\n').split()
                ret[chsum] = value

        except Exception as e:
            raise exception.ServiceUnavailable(e)

        if 'filesize' not in ret:
            raise exception.ServiceUnavailable('Filesize could not be retrieved.')
        if PREFERRED_CHECKSUM != chsum or not chsum:
            msg = '{} does not match with {}'.format(chsum, PREFERRED_CHECKSUM)
            raise exception.RSEChecksumUnavailable(msg)

        return ret

    def pfn2path(self, pfn):
        """
        Returns the path of a file given the pfn, i.e. scheme and hostname are subtracted from the pfn.

        :param path: pfn of a file

        :returns: path.
        """
        self.logger(logging.DEBUG, 'xrootd.pfn2path: pfn: {}'.format(pfn))
        if pfn.startswith('//'):
            return pfn
        elif pfn.startswith('/'):
            return '/' + pfn
        else:
            prefix = self.attributes['prefix']
            path = pfn.partition(self.attributes['prefix'])[2]
            path = prefix + path
            return path

    def lfns2pfns(self, lfns):
        """
        Returns a fully qualified PFN for the file referred by path.

        :param path: The path to the file.

        :returns: Fully qualified PFN.
        """
        self.logger(logging.DEBUG, 'xrootd.lfns2pfns: lfns: {}'.format(lfns))
        pfns = {}
        prefix = self.attributes['prefix']

        if not prefix.startswith('/'):
            prefix = ''.join(['/', prefix])
        if not prefix.endswith('/'):
            prefix = ''.join([prefix, '/'])

        lfns = [lfns] if isinstance(lfns, dict) else lfns
        for lfn in lfns:
            scope, name = lfn['scope'], lfn['name']
            if 'path' in lfn and lfn['path'] is not None:
                pfns['%s:%s' % (scope, name)] = ''.join([self.attributes['scheme'], '://', self.attributes['hostname'], ':', str(self.attributes['port']), prefix, lfn['path']])
            else:
                pfns['%s:%s' % (scope, name)] = ''.join([self.attributes['scheme'], '://', self.attributes['hostname'], ':', str(self.attributes['port']), prefix, self._get_path(scope=scope, name=name)])
        return pfns

    def connect(self):
        """ Establishes the actual connection to the referred RSE.

            :param credentials: Provides information to establish a connection
                to the referred storage system. For S3 connections these are
                access_key, secretkey, host_base, host_bucket, progress_meter
                and skip_existing.

            :raises RSEAccessDenied
        """
        self.logger(logging.DEBUG, 'xrootd.connect: port: {}, hostname {}'.format(self.port, self.hostname))
        try:
            # The query stats call is not implemented on some xroot doors.
            # Workaround: fail, if server does not reply within 10 seconds for static config query
            cmd = f'{self._auth_env} XRD_REQUESTTIMEOUT=10 xrdfs {self.hostname}:{self.port} query config {self.hostname}:{self.port}'
            self.logger(logging.DEBUG, 'xrootd.connect: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                raise exception.RSEAccessDenied(err)
        except Exception as e:
            raise exception.RSEAccessDenied(e)

    def close(self):
        """ Closes the connection to RSE."""
        pass

    def get(self, pfn, dest, transfer_timeout=None):
        """ Provides access to files stored inside connected the RSE.

            :param pfn: Physical file name of requested file
            :param dest: Name and path of the files when stored at the client
            :param transfer_timeout: Transfer timeout (in seconds) - dummy

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound
        """
        self.logger(logging.DEBUG, 'xrootd.get: pfn: {}'.format(pfn))
        try:
            cmd = f'{self._auth_env} xrdcp -f {pfn} {dest}'
            self.logger(logging.DEBUG, 'xrootd.get: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status == 54:
                raise exception.SourceNotFound()
            elif status != 0:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def put(self, filename, target, source_dir, transfer_timeout=None):
        """
            Allows to store files inside the referred RSE.

            :param source: path to the source file on the client file system
            :param target: path to the destination file on the storage
            :param source_dir: Path where the to be transferred files are stored in the local file system
            :param transfer_timeout: Transfer timeout (in seconds) - dummy

            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        self.logger(logging.DEBUG, 'xrootd.put: filename: {} target: {}'.format(filename, target))
        source_dir = source_dir or '.'
        source_url = '%s/%s' % (source_dir, filename)
        self.logger(logging.DEBUG, 'xrootd put: source url: {}'.format(source_url))
        path = self.path2pfn(target)
        if not os.path.exists(source_url):
            raise exception.SourceNotFound()
        try:
            cmd = f'{self._auth_env} xrdcp -f {source_url} {path}'
            self.logger(logging.DEBUG, 'xrootd.put: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def delete(self, pfn):
        """
            Deletes a file from the connected RSE.

            :param pfn: Physical file name

            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        self.logger(logging.DEBUG, 'xrootd.delete: pfn: {}'.format(pfn))
        if not self.exists(pfn):
            raise exception.SourceNotFound()
        try:
            path = self.pfn2path(pfn)
            cmd = f'{self._auth_env} xrdfs {self.hostname}:{self.port} rm {path}'
            self.logger(logging.DEBUG, 'xrootd.delete: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def rename(self, pfn, new_pfn):
        """ Allows to rename a file stored inside the connected RSE.

            :param pfn:      Current physical file name
            :param new_pfn  New physical file name
            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        self.logger(logging.DEBUG, 'xrootd.rename: pfn: {}'.format(pfn))
        if not self.exists(pfn):
            raise exception.SourceNotFound()
        try:
            path = self.pfn2path(pfn)
            new_path = self.pfn2path(new_pfn)
            new_dir = new_path[:new_path.rindex('/') + 1]
            cmd = f'{self._auth_env} xrdfs {self.hostname}:{self.port} mkdir -p {new_dir}'
            self.logger(logging.DEBUG, 'xrootd.stat: mkdir cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            cmd = f'{self._auth_env} xrdfs {self.hostname}:{self.port} mv {path} {new_path}'
            self.logger(logging.DEBUG, 'xrootd.stat: rename cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import TYPE_CHECKING, Any

from rucio.common.exception import RSENotFound
from rucio.common.types import InternalAccount, InternalScope
from rucio.core import permission
from rucio.core.rse import get_rse_id
from rucio.db.sqla.session import read_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from rucio.core.permission import PermissionResult


@read_session
def has_permission(issuer: str, action: str, kwargs: dict[str, Any], vo: str = 'def', *, session: "Session") -> 'PermissionResult':
    """
    Checks if an account has the specified permission to
    execute an action with parameters.

    :param issuer:  The Account issuer.
    :param vo:      The VO to check against.
    :param action:  The action (API call) called by the account.
    :param session: The db session to use
    :param kwargs:  List of arguments for the action.
    :returns: True if account is allowed to call the API call, otherwise False
    """

    kwargs = deepcopy(kwargs)
    if 'rse' in kwargs and 'rse_id' not in kwargs:
        try:
            rse_id = get_rse_id(rse=kwargs.get('rse'), vo=vo, session=session)
        except RSENotFound:
            rse_id = None
        kwargs.update({'rse_id': rse_id})

    if 'scope' in kwargs:
        kwargs['scope'] = InternalScope(kwargs['scope'], vo=vo)
    if 'attachments' in kwargs:
        for a in kwargs['attachments']:
            a['scope'] = InternalScope(a['scope'], vo=vo)

    if 'account' in kwargs:
        kwargs['account'] = InternalAccount(kwargs['account'], vo=vo)
    if 'accounts' in kwargs:
        kwargs['accounts'] = [InternalAccount(a, vo=vo) for a in kwargs['accounts']]
    if 'rules' in kwargs:
        for r in kwargs['rules']:
            r['account'] = InternalAccount(r['account'], vo=vo)
    if 'dids' in kwargs:
        for d in kwargs['dids']:
            d['scope'] = InternalScope(d['scope'], vo=vo)
            if 'rules' in d:
                for r in d['rules']:
                    r['account'] = InternalAccount(r['account'], vo=vo)

    issuer_account = InternalAccount(issuer, vo=vo)

    return permission.has_permission(issuer=issuer_account, action=action, kwargs=kwargs, session=session)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Download dumps via HTTP
"""

import collections
import datetime
import hashlib
import logging
import operator
import os
import re

from tabulate import tabulate

from rucio.common.dumper import DUMPS_CACHE_DIR, HTTPDownloadFailed, get_requests_session, http_download_to_file, smart_open, temp_file, to_datetime


class DataModel:
    """
    Data model for the dumps
    """

    BASE_URL = 'https://rucio-hadoop.cern.ch/'
    _FIELD_NAMES = None
    SCHEMA = []
    URI = None
    name = None

    def __init__(self, *args):
        if len(args) != len(self.SCHEMA):
            raise TypeError(
                'Wrong number of parameters (fields) to initialize {0} '
                'instance. {1} given, {2} expected:\n{3}'.format(
                    type(self).__name__,
                    len(args),
                    len(self.SCHEMA),
                    args,
                )
            )
        for (attr, parse), value in zip(self.SCHEMA, args):
            try:
                setattr(self, attr, parse(value))
            except ValueError as err:
                err.args = ('str parseable with {0} expected but got "{1}"'.format(
                    str(parse),
                    value,
                ),)
                raise err

        self.date = None

    @classmethod
    def get_fieldnames(cls) -> list[str]:
        """
        Get the field names
        """
        if cls._FIELD_NAMES is None:
            cls._FIELD_NAMES = [name for name, _ in cls.SCHEMA]
        return cls._FIELD_NAMES

    def pprint(self):
        """
        Pretty print the dump
        """
        return ''.join(
            ['{0}: {1}\n'.format(attr, getattr(self, attr)) for attr, _ in self.SCHEMA]
        )

    def __getitem__(self, index):
        """
        Return the item
        """
        return getattr(self, self.SCHEMA[index][0])

    @classmethod
    def csv_header(cls, fields=None):
        """
        Add the CSV header if necessary
        """
        if fields is None:
            fields = (field for field, _ in cls.SCHEMA)
        return ','.join(fields)

    def formated_fields(self, print_fields=None):
        """
        Reformat the fields
        """
        if print_fields is None:
            print_fields = (field for field, _ in self.SCHEMA)

        fields = []
        for attr in print_fields:
            field = getattr(self, attr)
            if isinstance(field, datetime.datetime):
                field = field.isoformat()
            fields.append(str(field))
        return fields

    def csv(self, fields=None):
        """
        Generate a CSV line
        """
        return ','.join(self.formated_fields(fields))

    @classmethod
    def tabulate_from(cls, iter_, format_='simple', fields=None):
        return tabulate(
            (row.formated_fields(fields) for row in iter_),
            (t[0] for t in cls.SCHEMA),
            format_,
        )

    @classmethod
    def each(cls, file, rse=None, date=None, filter_=None):
        if filter_ is None:
            filter_ = lambda x: True  # NOQA
        for line in file:
            record = cls.parse_line(line, rse, date)
            if filter_(record):
                yield record

    @classmethod
    def parse_line(cls, line, rse=None, date=None):
        fields = (field.strip() for field in line.split('\t'))
        instance = cls(*fields)
        instance.rse = rse
        instance.date = date
        return instance

    @classmethod
    def download(cls, rse, date='latest', cache_dir=DUMPS_CACHE_DIR):
        """
        Downloads the requested dump and returns an open read-only mode file
        like object.
        """
        logger = logging.getLogger('auditor.data_models')
        requests_session = get_requests_session()
        if date == 'latest':
            url = ''.join((cls.BASE_URL, cls.URI, '?rse={0}'.format(rse)))  # type: ignore
            request_headers = requests_session.head(url)
            for field in request_headers.headers['content-disposition'].split(';'):
                if field.startswith('filename='):
                    date = field.split('=')[1].split('_')[-1].split('.')[0]

        elif isinstance(date, datetime.datetime):
            date = date.strftime('%d-%m-%Y')
            url = ''.join((cls.BASE_URL, cls.URI, '?rse={0}&date={1}'.format(rse, date)))  # type: ignore
        else:
            raise ValueError("Passed date (%s) must be a datetime object or 'latest'." % date)

        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)

        filename = '{0}_{1}_{2}_{3}'.format(
            cls.__name__.lower(),
            rse,
            date,
            hashlib.sha1(url.encode()).hexdigest()
        )
        filename = re.sub(r'\W', '-', filename)
        path = os.path.join(cache_dir, filename)

        if not os.path.exists(path):
            logger.debug('Trying to download: "%s"', url)
            response = requests_session.head(url)
            if response.status_code != 200:
                logger.error(
                    'Retrieving %s returned %d status code',
                    url,
                    response.status_code,
                )
                raise HTTPDownloadFailed('Downloading {0} dump'.format(cls.__name__), code=str(response.status_code))

            with temp_file(cache_dir, final_name=filename) as (tfile, _):
                http_download_to_file(url, tfile, session=requests_session)

        return path

    @classmethod
    def dump(cls, rse, date='latest', filter_=None):
        filename = cls.download(rse, date)

        # Should check errors, content size at least
        file = smart_open(filename)

        return cls.each(file, rse, date, filter_)


class Dataset(DataModel):
    URI = 'datasets_per_rse'
    SCHEMA = (
        ('rse', str),
        ('scope', str),
        ('name', str),
        ('size', int),
        ('creation_date', to_datetime),
        ('update_date', to_datetime),
        ('last_access', to_datetime),
        ('state', str),
    )


class CompleteDataset(DataModel):
    URI = 'consistency_datasets'
    SCHEMA = (
        ('rse', str),
        ('scope', str),
        ('name', str),
        ('owner', str),
        ('size', lambda s: int(s) if s != '' else None),
        ('creation_date', to_datetime),
        ('last_access', to_datetime),
    )

    def __init__(self, *args):
        logger = logging.getLogger('auditor.data_models')
        super(CompleteDataset, self).__init__(*args[0:7])
        if len(args) == 8:
            logger.warning('Extra parameter\nrse: %s\ndataset: %s\n', self.rse, self.name)
            self.state = args[7]
        else:
            self.state = None
        if len(args) > 8:
            raise ValueError("Too many arguments, must be 8 or less. Instead passed %s" % len(args))


class Replica(DataModel):
    URI = 'replica_dumps'
    SCHEMA = (
        ('rse', str),
        ('scope', str),
        ('name', str),
        ('checksum', str),
        ('size', int),
        ('creation_date', to_datetime),
        ('path', str),
        ('update_date', to_datetime),
        ('state', str),
    )

    def __init__(self, *args):
        logger = logging.getLogger('auditor.data_models')
        if len(args) == 8:
            args = list(args)
            args.append(None)

        super(Replica, self).__init__(*args)

        if len(args) == 8:
            logger.warning('Missing parameter\nrse: %s\ndataset: %s\n', self.rse, self.name)
        elif len(args) > 9:
            raise ValueError("Too many arguments. Must be 9 or less, instead passed %s" % len(args))


class Filter:
    _Condition = collections.namedtuple('_Condition', ('comparator', 'attribute', 'expected'))

    def __init__(self, filter_str, record_class):
        '''
        Filter objects allow to match a DataModel subclass instance against
        one or more conditions.

        For the moment only equality conditions are implemented.

        :param filter_str: One or multiple comma separated conditions.
        :param record_class: DataModel subclass (used to check if the
        conditions use valid fields).

        Examples:
        available = Filter('state=A', Replica)
        available.match(replica)

        test_scope_avail = Filter('scope=test,state=A', Replica)
        test_scope_avail.match(replica)
        '''
        self.conditions = []
        for expr in filter_str.split(','):
            key, expected = expr.split('=')
            # Better checks required
            if key not in record_class.get_fieldnames():
                raise ValueError("Key %s not supported." % key)
            parser = list(filter(lambda t: t[0] == key, record_class.SCHEMA))[0][1]
            self.conditions.append(self._Condition(
                comparator=operator.eq,
                attribute=key,
                expected=parser(expected),
            ))

    def match(self, record):
        '''
        :param record: DataModel subclass instance.
        :returns: True if record matches all the conditions in this filter,
        else returns False.
        '''
        for cond in self.conditions:
            val = cond.comparator(
                getattr(record, cond.attribute),
                cond.expected,
            )

            if not val:
                return False

        return True



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from collections.abc import Sequence

    from rucio.common import types
    from rucio.core.rse import RseData
    from rucio.transfertool.transfertool import TransferStatusReport


class BittorrentDriver(metaclass=ABCMeta):
    external_name = ''
    required_rse_attrs = tuple()

    @classmethod
    @abstractmethod
    def make_driver(cls: "type[BittorrentDriver]", rse: "RseData", logger: 'types.LoggerFunction' = logging.log) -> "Optional[BittorrentDriver]":
        pass

    @abstractmethod
    def listen_addr(self) -> tuple[str, int]:
        pass

    @abstractmethod
    def add_torrent(self, file_name: str, file_content: bytes, download_location: str, seed_mode: bool = False) -> None:
        pass

    @abstractmethod
    def add_peers(self, torrent_id: str, peers: 'Sequence[tuple[str, int]]') -> None:
        pass

    @abstractmethod
    def get_status(self, request_id: str, torrent_id: str) -> "TransferStatusReport":
        pass



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import re
from configparser import NoOptionError, NoSectionError
from json import dumps
from typing import TYPE_CHECKING, Any, Optional

from sqlalchemy import and_, delete, func, select
from sqlalchemy.exc import IntegrityError, NoResultFound, StatementError
from sqlalchemy.orm import aliased

from rucio.common.config import config_get_bool
from rucio.common.exception import RucioException, SubscriptionDuplicate, SubscriptionNotFound
from rucio.db.sqla import models
from rucio.db.sqla.constants import RuleState, SubscriptionState
from rucio.db.sqla.session import read_session, stream_session, transactional_session

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from sqlalchemy.orm import Session

    from rucio.common.types import InternalAccount
    LoggerFunction = Callable[..., Any]
    SubscriptionType = dict


@transactional_session
def add_subscription(name: str,
                     account: "InternalAccount",
                     filter_: str,
                     replication_rules: str,
                     comments: str,
                     lifetime: Optional[int] = None,
                     retroactive: Optional[bool] = False,
                     dry_run: Optional[bool] = False,
                     priority: Optional[int] = 3,
                     *, session: "Session") -> str:
    """
    Adds a new subscription which will be verified against every new added file and dataset

    :param account:            Account identifier
    :type account:             String
    :param name:               Name of the subscription
    :type name:                String
    :param filter_:             Dictionary of attributes by which the input data should be filtered
                               **Example**: ``{'dsn': 'data11_hi*.express_express.*,data11_hi*physics_MinBiasOverlay*', 'account': 'tzero'}``
    :type filter_:              Dict
    :param replication_rules:  Replication rules to be set : Dictionary with keys copies, rse_expression, weight, rse_expression
    :type replication_rules:   Dict
    :param comments:           Comments for the subscription
    :type comments:            String
    :param lifetime:           Subscription's lifetime (days)
    :type lifetime:            Integer or None
    :param retroactive:        Flag to know if the subscription should be applied on previous data
    :type retroactive:         Boolean
    :param dry_run:            Just print the subscriptions actions without actually executing them (Useful if retroactive flag is set)
    :type dry_run:             Boolean
    :param priority:           The priority of the subscription
    :type priority:            Integer
    :param session:            The database session in use.

    :returns:                  The subscriptionid
    """
    try:
        keep_history = config_get_bool('subscriptions', 'keep_history')
    except (NoOptionError, NoSectionError, RuntimeError):
        keep_history = False

    SubscriptionHistory = models.SubscriptionHistory
    retroactive = bool(retroactive)  # Force boolean type, necessary for strict SQL
    state = SubscriptionState.ACTIVE
    if retroactive:
        state = SubscriptionState.NEW
    if lifetime:
        date_lifetime = datetime.datetime.utcnow() + datetime.timedelta(days=lifetime)
    else:
        date_lifetime = None
    new_subscription = models.Subscription(name=name,
                                           filter=filter_,
                                           account=account,
                                           replication_rules=replication_rules,
                                           state=state,
                                           lifetime=date_lifetime,
                                           retroactive=retroactive,
                                           policyid=priority, comments=comments)
    if keep_history:
        subscription_history = SubscriptionHistory(id=new_subscription.id,
                                                   name=new_subscription.name,
                                                   filter=new_subscription.filter,
                                                   account=new_subscription.account,
                                                   replication_rules=new_subscription.replication_rules,
                                                   state=new_subscription.state,
                                                   lifetime=new_subscription.lifetime,
                                                   retroactive=new_subscription.retroactive,
                                                   policyid=new_subscription.policyid,
                                                   comments=new_subscription.comments,
                                                   created_at=datetime.datetime.utcnow(),
                                                   updated_at=datetime.datetime.utcnow())
    try:
        new_subscription.save(session=session)
        if keep_history:
            subscription_history.save(session=session)
    except IntegrityError as error:
        if re.match('.*IntegrityError.*ORA-00001: unique constraint.*SUBSCRIPTIONS_PK.*violated.*', error.args[0])\
           or re.match(".*IntegrityError.*UNIQUE constraint failed: subscriptions.name, subscriptions.account.*", error.args[0])\
           or re.match('.*IntegrityError.*columns? name.*account.*not unique.*', error.args[0]) \
           or re.match('.*IntegrityError.*ORA-00001: unique constraint.*SUBSCRIPTIONS_NAME_ACCOUNT_UQ.*violated.*', error.args[0])\
           or re.match('.*IntegrityError.*1062.*Duplicate entry.*', error.args[0]) \
           or re.match('.*IntegrityError.*duplicate key value violates unique constraint.*', error.args[0]) \
           or re.match('.*UniqueViolation.*duplicate key value violates unique constraint.*', error.args[0]):
            raise SubscriptionDuplicate('Subscription \'%s\' owned by \'%s\' already exists!' % (name, account))
        raise RucioException(error.args)
    return str(new_subscription.id)


@transactional_session
def update_subscription(name: str,
                        account: "InternalAccount",
                        metadata: dict,
                        *, session: "Session") -> None:
    """
    Updates a subscription

    :param name:               Name of the subscription
    :type name:                String
    :param account:            Account identifier
    :type account:             String
    :param metadata:           Dictionary of metadata to update. Supported keys : filter, replication_rules, comments, lifetime, retroactive, dry_run, priority, last_processed
    :type metadata:            Dict
    :param session:            The database session in use.

    :raises: SubscriptionNotFound if subscription is not found
    """
    try:
        keep_history = config_get_bool('subscriptions', 'keep_history')
    except (NoOptionError, NoSectionError, RuntimeError):
        keep_history = False
    values: dict[str, Any] = {'state': SubscriptionState.UPDATED}
    if 'filter' in metadata and metadata['filter']:
        values['filter'] = dumps(metadata['filter'])
    if 'replication_rules' in metadata and metadata['replication_rules']:
        values['replication_rules'] = dumps(metadata['replication_rules'])
    if 'lifetime' in metadata and metadata['lifetime']:
        values['lifetime'] = datetime.datetime.utcnow() + datetime.timedelta(days=float(metadata['lifetime']))
    if 'retroactive' in metadata and metadata['retroactive']:
        values['retroactive'] = metadata['retroactive']
    if 'dry_run' in metadata and metadata['dry_run']:
        values['dry_run'] = metadata['dry_run']
    if 'comments' in metadata and metadata['comments']:
        values['comments'] = metadata['comments']
    if 'priority' in metadata and metadata['priority']:
        values['policyid'] = metadata['priority']
    if 'last_processed' in metadata and metadata['last_processed']:
        values['last_processed'] = metadata['last_processed']
    if 'state' in metadata and metadata['state'] == SubscriptionState.INACTIVE:
        values['state'] = SubscriptionState.INACTIVE
        values['expired_at'] = datetime.datetime.utcnow()

    SubscriptionHistory = models.SubscriptionHistory
    try:
        stmt = select(
            models.Subscription
        ).where(
            and_(models.Subscription.name == name,
                 models.Subscription.account == account)
        )
        subscription = session.execute(stmt).scalar_one()

        # To avoid update in the subscription history table whenever last processed field is changed
        current_subscription_state = subscription.to_dict()
        new_subscription_state = values.copy()

        for key in ["updated_at", "last_processed"]:
            new_subscription_state.pop(key, "None")
            current_subscription_state.pop(key, "None")

        subscription.update(values)
        if keep_history and current_subscription_state != new_subscription_state:
            subscription_history = SubscriptionHistory(id=subscription.id,
                                                       name=subscription.name,
                                                       filter=subscription.filter,
                                                       account=subscription.account,
                                                       replication_rules=subscription.replication_rules,
                                                       state=subscription.state,
                                                       lifetime=subscription.lifetime,
                                                       retroactive=subscription.retroactive,
                                                       policyid=subscription.policyid,
                                                       comments=subscription.comments,
                                                       last_processed=subscription.last_processed,
                                                       expired_at=subscription.expired_at,
                                                       updated_at=datetime.datetime.utcnow(),
                                                       created_at=subscription.created_at)
            subscription_history.save(session=session)
    except NoResultFound:
        raise SubscriptionNotFound(f"Subscription for account '{account}' named '{name}' not found")


@stream_session
def list_subscriptions(name: Optional[str] = None,
                       account: "Optional[InternalAccount]" = None,
                       state: Optional[SubscriptionState] = None,
                       *, session: "Session",
                       logger: "LoggerFunction" = logging.log) -> "Iterator[SubscriptionType]":
    """
    Returns a dictionary with the subscription information :
    Examples: ``{'status': 'INACTIVE/ACTIVE/BROKEN', 'last_modified_date': ...}``

    :param name:               Name of the subscription
    :type name:                String
    :param account:            Account identifier
    :type account:             String
    :param session:            The database session in use.
    :param logger:             Optional decorated logger that can be passed from the calling daemons or servers.

    :returns:                  Dictionary containing subscription parameter
    :rtype:                    Dict
    :raises:                   exception.NotFound if subscription is not found
    """
    stmt = select(
        models.Subscription
    )
    try:
        if name:
            stmt = stmt.where(
                models.Subscription.name == name
            )
        if account:
            if account.internal is not None and '*' in account.internal:
                account_str = account.internal.replace('*', '%')
                stmt = stmt.where(
                    models.Subscription.account.like(account_str)
                )
            else:
                stmt = stmt.where(
                    models.Subscription.account == account
                )
        if state:
            stmt = stmt.where(
                models.Subscription.state == state
            )
    except IntegrityError as error:
        logger(logging.ERROR, str(error))
        raise RucioException(error.args)
    found = False
    for row in session.execute(stmt).scalars().all():
        found = True
        yield row.to_dict()
    if not found:
        raise SubscriptionNotFound(f"Subscription for account '{account}' named '{name}' not found")


@transactional_session
def delete_subscription(subscription_id: str, *, session: "Session") -> None:
    """
    Deletes a subscription

    :param subscription_id: Subscription identifier
    :type subscription_id:  String
    """
    stmt = delete(
        models.Subscription
    ).where(
        models.Subscription.id == subscription_id
    )
    session.execute(stmt)


@stream_session
def list_subscription_rule_states(
    name: Optional[str] = None,
    account: Optional["InternalAccount"] = None,
    *,
    session: "Session",
    logger: "LoggerFunction" = logging.log
) -> "Iterator[tuple[InternalAccount, str, RuleState, int]]":
    """Returns a list of with the number of rules per state for a subscription.

    :param name:               Name of the subscription
    :param account:            Account identifier
    :param session:            The database session in use.
    :param logger:             Optional decorated logger that can be passed from the calling daemons or servers.
    :returns:                  List with tuple (account, name, state, count)
    """
    subscription = aliased(models.Subscription)
    rule = aliased(models.ReplicationRule)
    # count needs a label to allow conversion to dict (label name can be changed)
    stmt = select(
        subscription.account,
        subscription.name,
        rule.state,
        func.count().label('count')
    ).join(
        rule,
        subscription.id == rule.subscription_id
    )

    try:
        if name:
            stmt = stmt.where(
                subscription.name == name
            )

        if account:
            if account.internal and '*' in account.internal:
                account_str = account.internal.replace('*', '%')
                stmt = stmt.where(
                    subscription.account.like(account_str)
                )
            else:
                stmt = stmt.where(
                    subscription.account == account
                )

    except IntegrityError as error:
        logger(logging.ERROR, str(error))
        raise RucioException(error.args)

    stmt = stmt.group_by(
        subscription.account,
        subscription.name,
        rule.state
    )

    for row in session.execute(stmt).all():
        yield row._tuple()


@read_session
def get_subscription_by_id(subscription_id: str, *, session: "Session") -> "SubscriptionType":
    """
    Get a specific subscription by id.

    :param subscription_id:    The subscription_id to select.
    :param session:            The database session in use.

    :raises:                   SubscriptionNotFound if no Subscription can be found.
    """

    try:
        stmt = select(
            models.Subscription
        ).where(
            models.Subscription.id == subscription_id
        )
        return session.execute(stmt).scalar_one().to_dict()
    except NoResultFound:
        raise SubscriptionNotFound('No subscription with the id %s found' % (subscription_id))
    except StatementError:
        raise RucioException('Badly formatted subscription id (%s)' % (subscription_id))



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from dogpile.cache.region import CacheRegion

from rucio.common.client import is_client
from rucio.common.config import config_get

if TYPE_CHECKING:
    from collections.abc import Callable


CACHE_URL = config_get('cache', 'url', False, '127.0.0.1:11211', check_config_table=False)

ENABLE_CACHING = True
_mc_client = None
try:
    if is_client():
        ENABLE_CACHING = False
    else:
        import pymemcache
        _mc_client = pymemcache.Client(CACHE_URL, connect_timeout=1, timeout=1)
        _mc_client.version()
except OSError:
    ENABLE_CACHING = False
except ImportError:
    ENABLE_CACHING = False
finally:
    if _mc_client:
        _mc_client.close()


class MemcacheRegion(CacheRegion):
    """
    Subclass of CacheRegion.
    It uses pymemcache as backend if ENABLE_CACHING is True,
    otherwise it it configured to null.
    """
    def __init__(
            self,
            expiration_time: int,
            function_key_generator: Optional['Callable'] = None,
            memcached_expire_time: Optional[int] = None
    ):
        if function_key_generator:
            super().__init__(function_key_generator=function_key_generator)
        else:
            super().__init__()
        self._configure_region(expiration_time, memcached_expire_time)

    def _configure_region(
            self,
            expiration_time: int,
            memcached_expire_time: Optional[int]
    ) -> None:
        if ENABLE_CACHING:
            self.configure(
                'dogpile.cache.pymemcache',
                expiration_time=expiration_time,
                arguments={
                    'url': CACHE_URL,
                    'distributed_lock': True,
                    'memcached_expire_time': memcached_expire_time if memcached_expire_time else expiration_time + 60,  # must be bigger than expiration_time
                }
            )
        else:
            self.configure('dogpile.cache.null')


class CacheKey:
    """
    Helper class to generate cache keys
    based on sections and options.
    """

    @staticmethod
    def _generate_key(*args: str) -> str:
        return '_'.join(args)

    @staticmethod
    def has_section(section: str) -> str:
        return CacheKey._generate_key('has_section', section)

    @staticmethod
    def options(section: str) -> str:
        return CacheKey._generate_key('options', section)

    @staticmethod
    def has_option(section: str, option: str) -> str:
        return CacheKey._generate_key('has_option', section, option)

    @staticmethod
    def items(section: str) -> str:
        return CacheKey._generate_key('items', section)

    @staticmethod
    def value(section: str, option: str) -> str:
        return CacheKey._generate_key('get', section, option)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from enum import Enum
from re import match
from traceback import format_exc
from typing import TYPE_CHECKING, Any, Optional

from sqlalchemy import and_, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import exc

import rucio.core.account_counter
import rucio.core.rse
from rucio.common import exception
from rucio.common.config import config_get_bool
from rucio.core.vo import vo_exists
from rucio.db.sqla import models
from rucio.db.sqla.constants import AccountStatus, AccountType
from rucio.db.sqla.session import read_session, stream_session, transactional_session

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from sqlalchemy.orm import Session

    from rucio.common.types import AccountAttributesDict, AccountDict, AccountUsageModelDict, IdentityDict, InternalAccount, UsageDict


@transactional_session
def add_account(
    account: "InternalAccount",
    type_: AccountType,
    email: str,
    *,
    session: "Session"
) -> None:
    """ Add an account with the given account name and type.

    :param account: the name of the new account.
    :param type_: the type of the new account.
    :param email: The Email address associated with the account.
    :param session: the database session in use.
    """
    vo = account.vo
    if not vo_exists(vo=vo, session=session):
        raise exception.VONotFound('VO {} not found'.format(vo))

    # Reserve the name 'super_root' for multi_vo admins
    if account.external == 'super_root':
        if not (vo == 'def' and config_get_bool('common', 'multi_vo', raise_exception=False, default=False)):
            raise exception.UnsupportedAccountName('The name "%s" cannot be used.' % account.external)

    new_account = models.Account(account=account, account_type=type_, email=email,
                                 status=AccountStatus.ACTIVE)
    try:
        new_account.save(session=session)
    except IntegrityError:
        raise exception.Duplicate('Account ID \'%s\' already exists!' % account)


@read_session
def account_exists(
    account: "InternalAccount",
    *,
    session: "Session"
) -> bool:
    """ Checks to see if account exists and is active.

    :param account: Name of the account.
    :param session: the database session in use.

    :returns: True if found, otherwise false.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account,
        models.Account.status == AccountStatus.ACTIVE
    )
    return session.execute(query).scalar() is not None


@read_session
def get_account(
    account: "InternalAccount",
    *,
    session: "Session"
) -> models.Account:
    """ Returns an account for the given account name.

    :param account: the name of the account.
    :param session: the database session in use.

    :returns: a dict with all information for the account.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account
    )

    result = session.execute(query).scalar()
    if result is None:
        raise exception.AccountNotFound('Account with ID \'%s\' cannot be found' % account)
    return result


@transactional_session
def del_account(
    account: "InternalAccount",
    *,
    session: "Session"
) -> None:
    """ Disable an account with the given account name.

    :param account: the account name.
    :param session: the database session in use.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account,
        models.Account.status == AccountStatus.ACTIVE
    )
    try:
        query_result = session.execute(query).scalar_one()
    except exc.NoResultFound:
        raise exception.AccountNotFound('Account with ID \'%s\' cannot be found' % account)

    query_result.update({'status': AccountStatus.DELETED, 'deleted_at': datetime.utcnow()})


@transactional_session
def update_account(
    account: "InternalAccount",
    key: str,
    value: Any,
    *,
    session: "Session"
) -> None:
    """ Update a property of an account.

    :param account: Name of the account.
    :param key: Account property like status.
    :param value: Property value.
    :param session: the database session in use.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account
    )
    try:
        query_result = session.execute(query).scalar_one()
    except exc.NoResultFound:
        raise exception.AccountNotFound('Account with ID \'%s\' cannot be found' % account)
    if key == 'status':
        if isinstance(value, str):
            value = AccountStatus[value]
        if value == AccountStatus.SUSPENDED:
            query_result.update({'status': value, 'suspended_at': datetime.utcnow()})
        elif value == AccountStatus.ACTIVE:
            query_result.update({'status': value, 'suspended_at': None})
    else:
        query_result.update({key: value})


@stream_session
def list_accounts(
    filter_: Optional["Mapping[str, Any]"] = None,
    *,
    session: "Session"
) -> "Iterator[AccountDict]":
    """ Returns a list of all account names.

    :param filter_: Dictionary of attributes by which the input data should be filtered
    :param session: the database session in use.

    returns: a list of all account names.
    """
    if filter_ is None:
        filter_ = {}
    query = select(
        models.Account.account,
        models.Account.account_type,
        models.Account.email
    ).where(
        models.Account.status == AccountStatus.ACTIVE
    )
    for filter_type in filter_:
        if filter_type == 'account_type':
            if isinstance(filter_['account_type'], str):
                query = query.where(
                    models.Account.account_type == AccountType[filter_['account_type']]
                )
            elif isinstance(filter_['account_type'], Enum):
                query = query.where(
                    models.Account.account_type == filter_['account_type']
                )

        elif filter_type == 'identity':
            query = query.join(
                models.IdentityAccountAssociation,
                models.Account.account == models.IdentityAccountAssociation.account
            ).where(
                models.IdentityAccountAssociation.identity == filter_['identity']
            )

        elif filter_type == 'account':
            if '*' in filter_['account'].internal:
                account_str = filter_['account'].internal.replace('*', '%')
                query = query.where(
                    models.Account.account.like(account_str)
                )
            else:
                query = query.where(
                    models.Account.account == filter_['account']
                )
        else:
            query = query.join(
                models.AccountAttrAssociation,
                models.Account.account == models.AccountAttrAssociation.account
            ).where(
                models.AccountAttrAssociation.key == filter_type,
                models.AccountAttrAssociation.value == filter_[filter_type]
            )
    query = query.order_by(models.Account.account)

    for account, account_type, email in session.execute(query).yield_per(25):
        yield {'account': account, 'type': account_type, 'email': email}


@read_session
def list_identities(
    account: "InternalAccount",
    *,
    session: "Session"
) -> list["IdentityDict"]:
    """
    List all identities on an account.

    :param account: The account name.
    :param session: the database session in use.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account,
        models.Account.status == AccountStatus.ACTIVE
    )
    try:
        session.execute(query).scalar_one()
    except exc.NoResultFound:
        raise exception.AccountNotFound('Account with ID \'%s\' cannot be found' % account)

    query = select(
        models.IdentityAccountAssociation.identity_type.label('type'),
        models.IdentityAccountAssociation.identity,
        models.Identity.email
    ).join(
        models.Identity,
        and_(
            models.Identity.identity == models.IdentityAccountAssociation.identity,
            models.Identity.identity_type == models.IdentityAccountAssociation.identity_type
        )
    ).where(
        models.IdentityAccountAssociation.account == account
    )

    return [row._asdict() for row in session.execute(query)]  # type: ignore (pending SQLA2.1: https://github.com/rucio/rucio/discussions/6615)


@read_session
def list_account_attributes(
    account: "InternalAccount",
    *,
    session: "Session"
) -> list["AccountAttributesDict"]:
    """
    Get all attributes defined for an account.

    :param account: the account name to list the scopes of.
    :param session: The database session in use.

    :returns: a list of all key, value pairs for this account.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account,
        models.Account.status == AccountStatus.ACTIVE
    )
    try:
        session.execute(query).scalar_one()
    except exc.NoResultFound:
        raise exception.AccountNotFound("Account ID '{0}' does not exist".format(account))

    query = select(
        models.AccountAttrAssociation.key,
        models.AccountAttrAssociation.value
    ).where(
        models.AccountAttrAssociation.account == account
    )
    return [row._asdict() for row in session.execute(query)]  # type: ignore (pending SQLA2.1: https://github.com/rucio/rucio/discussions/6615)


@read_session
def has_account_attribute(
    account: "InternalAccount",
    key: str,
    *,
    session: "Session"
) -> bool:
    """
    Indicates whether the named key is present for the account.

    :param account: the account name to list the scopes of.
    :param key: the key for the attribute.
    :param session: The database session in use.

    :returns: True or False
    """
    query = select(
        models.AccountAttrAssociation.value
    ).where(
        models.AccountAttrAssociation.account == account,
        models.AccountAttrAssociation.key == key
    )
    return session.execute(query).scalar() is not None


@transactional_session
def add_account_attribute(
    account: "InternalAccount",
    key: str,
    value: Any,
    *,
    session: "Session"
) -> None:
    """
    Add an attribute for the given account name.

    :param key: the key for the new attribute.
    :param value: the value for the new attribute.
    :param account: the account to add the attribute to.
    :param session: The database session in use.
    """
    query = select(
        models.Account
    ).where(
        models.Account.account == account,
        models.Account.status == AccountStatus.ACTIVE
    )
    try:
        session.execute(query).scalar_one()
    except exc.NoResultFound:
        raise exception.AccountNotFound("Account ID '{0}' does not exist".format(account))

    new_attr = models.AccountAttrAssociation(account=account, key=key, value=value)
    try:
        new_attr.save(session=session)
    except IntegrityError as error:
        if match('.*IntegrityError.*ORA-00001: unique constraint.*ACCOUNT_ATTR_MAP_PK.*violated.*', error.args[0]) \
           or match('.*IntegrityError.*1062.*Duplicate entry.*for key.*', error.args[0]) \
           or match('.*IntegrityError.*UNIQUE constraint failed: account_attr_map.account, account_attr_map.key.*', error.args[0]) \
           or match('.*IntegrityError.*columns? account.*key.*not unique.*', error.args[0]) \
           or match('.*IntegrityError.*duplicate key value violates unique constraint.*', error.args[0]) \
           or match('.*UniqueViolation.*duplicate key value violates unique constraint.*', error.args[0]):
            raise exception.Duplicate('Key {0} already exist for account {1}!'.format(key, account))
    except Exception:
        raise exception.RucioException(str(format_exc()))


@transactional_session
def del_account_attribute(
    account: "InternalAccount",
    key: str,
    *,
    session: "Session"
) -> None:
    """
    Add an attribute for the given account name.

    :param account: the account to add the attribute to.
    :param key: the key for the new attribute.
    :param session: The database session in use.
    """
    query = select(
        models.AccountAttrAssociation
    ).where(
        models.AccountAttrAssociation.account == account,
        models.AccountAttrAssociation.key == key
    )
    aid = session.execute(query).scalar()
    if aid is None:
        raise exception.AccountNotFound('Attribute ({0}) does not exist for the account {1}!'.format(key, account))
    aid.delete(session=session)


@read_session
def get_usage(
    rse_id: str,
    account: "InternalAccount",
    *,
    session: "Session"
) -> "UsageDict":
    """
    Returns current values of the specified counter, or raises CounterNotFound if the counter does not exist.

    :param rse_id:           The id of the RSE.
    :param account:          The account name.
    :param session:          The database session in use.
    :returns:                A dictionary {'bytes', 'files', 'updated_at'}
    """
    query = select(
        models.AccountUsage.bytes,
        models.AccountUsage.files,
        models.AccountUsage.updated_at
    ).where(
        models.AccountUsage.rse_id == rse_id,
        models.AccountUsage.account == account
    )
    try:
        return session.execute(query).one()._asdict()  # type: ignore (pending SQLA2.1: https://github.com/rucio/rucio/discussions/6615)
    except exc.NoResultFound:
        return {'bytes': 0, 'files': 0, 'updated_at': None}


@read_session
def get_all_rse_usages_per_account(
    account: "InternalAccount",
    *,
    session: "Session"
) -> list["AccountUsageModelDict"]:
    """
    Returns current values of the specified counter, or raises CounterNotFound if the counter does not exist.

    :param rse_id:           The id of the RSE.
    :param account:          The account name.
    :param session:          The database session in use.
    :returns:                List of dicts with :py:class:`models.AccountUsage` items
    """
    query = select(
        models.AccountUsage
    ).where(
        models.AccountUsage.account == account
    )
    try:
        return [result.to_dict() for result in session.execute(query).scalars()]  # type: ignore (pending SQLA2.1: https://github.com/rucio/rucio/discussions/6615)
    except exc.NoResultFound:
        return []


@read_session
def get_usage_history(
    rse_id: str,
    account: "InternalAccount",
    *,
    session: "Session"
) -> list["UsageDict"]:
    """
    Returns historical values of the specified counter, or raises CounterNotFound if the counter does not exist.

    :param rse_id:           The id of the RSE.
    :param account:          The account name.
    :param session:          The database session in use.
    :returns:                A dictionary {'bytes', 'files', 'updated_at'}
    """
    query = select(
        models.AccountUsageHistory.bytes,
        models.AccountUsageHistory.files,
        models.AccountUsageHistory.updated_at
    ).where(
        models.AccountUsageHistory.rse_id == rse_id,
        models.AccountUsageHistory.account == account
    ).order_by(
        models.AccountUsageHistory.updated_at
    )
    try:
        return [row._asdict() for row in session.execute(query)]  # type: ignore (pending SQLA2.1: https://github.com/rucio/rucio/discussions/6615)
    except exc.NoResultFound:
        raise exception.CounterNotFound('No usage can be found for account %s on RSE %s' % (account, rucio.core.rse.get_rse_name(rse_id=rse_id, session=session)))
    return []



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import os

from rucio.common.exception import FileAlreadyExists, ServiceUnavailable, SourceNotFound
from rucio.rse.protocols import protocol

try:
    import arc  # pylint: disable=import-error
except:
    pass


class DataPoint:
    '''
    Wrapper around arc.datapoint_from_url() which does not clean up DataPoints
    when python objects are destroyed, leading to connection leaking when used
    with gridftp. This class should be used instead of arc.datapoint_from_url().
    It can be called like dp = DataPoint('gsiftp://...', uc); dp.h.Stat()
    where uc is an arc.UserConfig object.
    '''
    def __init__(self, u, uc):
        self.h = arc.datapoint_from_url(u, uc)

    def __del__(self):
        arc.DataPoint.__swig_destroy__(self.h)


class Default(protocol.RSEProtocol):
    """ Implementing access to RSEs using ARC client."""

    def __init__(self, protocol_attr, rse_settings, logger=None):
        """
        Set up UserConfig object.
        """
        super(Default, self).__init__(protocol_attr, rse_settings, logger=logger)

        # Arc logging to stdout, uncomment for debugging. Should use root
        # logger level eventually.
#         root_logger = arc.Logger_getRootLogger()
#         stream = arc.LogStream(sys.stdout)
#         root_logger.addDestination(stream)
#         # Set threshold to VERBOSE or DEBUG for more information
#         root_logger.setThreshold(arc.DEBUG)

        self.cfg = arc.UserConfig()
        try:
            self.cfg.ProxyPath(os.environ['X509_USER_PROXY'])
        except:
            pass

    def path2pfn(self, path):
        """
            Returns a fully qualified PFN for the file referred by path.

            :param path: The path to the file.

            :returns: Fully qualified PFN.

        """
        return ''.join([self.rse['scheme'], '://%s' % self.rse['hostname'], path])

    def exists(self, pfn):
        """ Checks if the requested file is known by the referred RSE.

            :param pfn: Physical file name

            :returns: True if the file exists, False if it doesn't

            :raise  ServiceUnavailable
        """
        dp = DataPoint(str(pfn), self.cfg)
        fileinfo = arc.FileInfo()

        status = dp.h.Stat(fileinfo)
        if not status:
            if status.GetErrno() == errno.ENOENT:
                return False
            raise ServiceUnavailable(str(status))

        return True

    def connect(self):
        """ Establishes the actual connection to the referred RSE.

            :raise RSEAccessDenied
        """
        pass

    def close(self):
        """ Closes the connection to RSE."""
        pass

    def __arc_copy(self, src, dest, space_token=None, transfer_timeout=None):

        # TODO set proxy path

        # Convert the arguments to DataPoint objects
        source = DataPoint(str(src), self.cfg)
        if source.h is None:
            raise ServiceUnavailable("Can't handle source %s" % src)

        destination = DataPoint(str(dest), self.cfg)
        if destination.h is None:
            raise ServiceUnavailable("Can't handle destination %s" % dest)
        if space_token:
            destination.h.GetURL().AddOption('spacetoken', space_token)

        # DataMover does the transfer
        mover = arc.DataMover()
        # Don't attempt to retry on error
        mover.retry(False)
        # Passive and insecure gridftp
        mover.passive(True)
        mover.secure(False)
        # Do the transfer
        status = mover.Transfer(source.h, destination.h, arc.FileCache(), arc.URLMap())

        if not status:
            if status.GetErrno() == errno.ENOENT:
                raise SourceNotFound()
            if status.GetErrno() == errno.EEXIST:
                raise FileAlreadyExists()
            raise ServiceUnavailable(str(status))

    def get(self, pfn, dest, transfer_timeout=None):
        """ Provides access to files stored inside connected the RSE.

            :param pfn: Physical file name of requested file
            :param dest: Name and path of the files when stored at the client
            :param transfer_timeout Transfer timeout (in seconds) - dummy

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound
        """
        self.__arc_copy(pfn, dest, transfer_timeout=transfer_timeout)

    def put(self, source, target, source_dir=None, transfer_timeout=None):
        """ Allows to store files inside the referred RSE.

            :param source: Physical file name
            :param target: Name of the file on the storage system e.g. with prefixed scope
            :param source_dir Path where the to be transferred files are stored in the local file system
            :param transfer_timeout Transfer timeout (in seconds) - dummy

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound
        """

        if source_dir:
            sf = source_dir + '/' + source
        else:
            sf = source

        space_token = None
        if self.attributes['extended_attributes'] is not None and 'space_token' in list(self.attributes['extended_attributes'].keys()):
            space_token = self.attributes['extended_attributes']['space_token']

        self.__arc_copy(sf, target, space_token, transfer_timeout=transfer_timeout)

    def delete(self, pfn):
        """ Deletes a file from the connected RSE.

            :param pfn: Physical file name

            :raises ServiceUnavailable, SourceNotFound
        """
        dp = DataPoint(str(pfn), self.cfg)
        if dp.h is None:
            raise ServiceUnavailable("Can't handle pfn %s" % pfn)

        status = dp.h.Remove()
        if not status:
            if status.GetErrno() == errno.ENOENT:
                raise SourceNotFound()
            raise ServiceUnavailable(str(status))

    def rename(self, pfn, new_pfn):
        """ Allows to rename a file stored inside the connected RSE.

            :param pfn:      Current physical file name
            :param new_pfn  New physical file name

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound
        """
        dp = DataPoint(str(pfn), self.cfg)
        if dp.h is None:
            raise ServiceUnavailable("Can't handle pfn %s" % pfn)

        url = arc.URL(str(new_pfn))
        if not url:
            raise ServiceUnavailable("Can't handle new pfn %s" % new_pfn)

        status = dp.h.Rename(url)
        if not status:
            if status.GetErrno() == errno.ENOENT:
                raise SourceNotFound()
            raise ServiceUnavailable(str(status))



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import TYPE_CHECKING, Any, Optional

from sqlalchemy import and_, delete, or_, select
from sqlalchemy.sql.expression import false, insert

from rucio.common.utils import chunks
from rucio.db.sqla import filter_thread_work, models
from rucio.db.sqla.session import read_session, transactional_session

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sqlalchemy.orm import Session


@transactional_session
def add_quarantined_replicas(rse_id: str, replicas: list[dict[str, Any]], *, session: "Session") -> None:
    """
    Bulk add quarantined file replicas.

    :param rse_id:      The rse id.
    :param replicas: A list of dicts with the replica information.
    :param session:  The database session in use.
    """

    # Exclude files that have a registered replica.  This is a
    # safeguard against potential issues in the Auditor.
    file_clause = []
    for replica in replicas:
        if "scope" in replica and "name" in replica:
            file_clause.append(and_(models.RSEFileAssociation.scope == replica['scope'],
                                    models.RSEFileAssociation.name == replica['name'],
                                    models.RSEFileAssociation.rse_id == rse_id))
    if file_clause:
        stmt = select(
            models.RSEFileAssociation.scope,
            models.RSEFileAssociation.name,
            models.RSEFileAssociation.rse_id
        ).with_hint(
            models.RSEFileAssociation,
            'INDEX(REPLICAS REPLICAS_PK)',
            'oracle'
        ).where(
            or_(*file_clause)
        )
        existing_replicas = [(scope, name, rseid) for scope, name, rseid in session.execute(stmt).all()]
        replicas = [replica for replica in replicas if (replica.get('scope', None), replica.get('name', None), rse_id) not in existing_replicas]

    # Exclude files that have already been added to the quarantined
    # replica table.
    quarantine_clause = []
    for replica in replicas:
        quarantine_clause.append(and_(models.QuarantinedReplica.path == replica['path'],
                                      models.QuarantinedReplica.rse_id == rse_id))
    stmt = select(
        models.QuarantinedReplica.path,
        models.QuarantinedReplica.rse_id
    ).where(
        or_(*quarantine_clause)
    )
    quarantine_replicas = [(path, rseid) for path, rseid in session.execute(stmt).all()]
    replicas = [replica for replica in replicas if (replica['path'], rse_id) not in quarantine_replicas]

    values = [{'rse_id': rse_id,
               'path': replica['path'],
               'scope': replica.get('scope'),
               'name': replica.get('name'),
               'bytes': replica.get('bytes')}
              for replica in replicas]
    stmt = insert(
        models.QuarantinedReplica
    )
    session.execute(stmt, values)


@transactional_session
def delete_quarantined_replicas(rse_id: str, replicas: "Iterable[dict[str, Any]]", *, session: "Session") -> None:
    """
    Delete file replicas.

    :param rse_id: the rse id.
    :param replicas: An iterable of dicts with the replica information.
    :param session: The database session in use.
    """

    conditions = []
    for replica in replicas:
        conditions.append(models.QuarantinedReplica.path == replica['path'])

    if conditions:
        stmt = delete(
            models.QuarantinedReplica
        ).where(
            and_(models.QuarantinedReplica.rse_id == rse_id,
                 or_(*conditions))
        ).execution_options(
            synchronize_session=False
        )
        session.execute(stmt)

    if replicas:
        values = [{'rse_id': rse_id, 'path': replica['path'],
                   'bytes': replica.get('bytes'),
                   'created_at': replica.get('created_at'),
                   'deleted_at': datetime.datetime.utcnow()}
                  for replica in replicas]
        stmt = insert(
            models.QuarantinedReplicaHistory
        )
        session.execute(stmt, values)


@read_session
def list_quarantined_replicas(rse_id: str, limit: int, worker_number: Optional[int] = None, total_workers: Optional[int] = None, *, session: "Session") -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
    """
    List RSE Quarantined File replicas.

    :param rse_id: the rse id.
    :param limit: The maximum number of replicas returned.
    :param worker_number:      id of the executing worker.
    :param total_workers:      Number of total workers.
    :param session: The database session in use.

    :returns: two lists :
              - The first one contains quarantine replicas actually registered in the replicas tables
              - The second one contains real "dark" files
    """

    replicas_clause = []
    quarantined_replicas = {}
    real_replicas = []
    dark_replicas = []
    stmt = select(
        models.QuarantinedReplica.path,
        models.QuarantinedReplica.bytes,
        models.QuarantinedReplica.scope,
        models.QuarantinedReplica.name,
        models.QuarantinedReplica.created_at
    ).where(
        models.QuarantinedReplica.rse_id == rse_id
    )
    stmt = filter_thread_work(session=session, query=stmt, total_threads=total_workers, thread_id=worker_number, hash_variable='path')
    stmt = stmt.limit(
        limit
    )
    for path, bytes_, scope, name, created_at in session.execute(stmt).all():
        if not (scope, name) in quarantined_replicas:
            quarantined_replicas[(scope, name)] = []
            replicas_clause.append(and_(models.RSEFileAssociation.scope == scope,
                                        models.RSEFileAssociation.name == name))
        quarantined_replicas[(scope, name)].append((path, bytes_, created_at))

    stmt = select(
        models.RSEFileAssociation.scope,
        models.RSEFileAssociation.name
    ).where(
        models.RSEFileAssociation.rse_id == rse_id
    )
    for chunk in chunks(replicas_clause, 20):
        curr_stmt = stmt.where(
            or_(*chunk)
        )

        for scope, name in session.execute(curr_stmt).all():
            reps = quarantined_replicas.pop((scope, name))
            real_replicas.extend([{'scope': scope,
                                   'name': name,
                                   'rse_id': rse_id,
                                   'path': rep[0],
                                   'bytes': rep[1],
                                   'created_at': rep[2]}
                                  for rep in reps])

    for key, value in quarantined_replicas.items():
        dark_replicas.extend([{'scope': key[0],
                               'name': key[1],
                               'rse_id': rse_id,
                               'path': rep[0],
                               'bytes': rep[1],
                               'created_at': rep[2]}
                              for rep in value])

    return real_replicas, dark_replicas


@read_session
def list_rses_with_quarantined_replicas(filters: Optional[dict[str, Any]] = None, *, session: "Session") -> list[str]:
    """
    List RSEs in the Quarantined Queues.

    :param filters: dictionary of attributes by which the results should be filtered.
    :param session: The database session in use.

    :returns: a list of RSEs.
    """
    stmt = select(
        models.RSE.id
    ).distinct(
    ).where(
        and_(models.QuarantinedReplica.rse_id == models.RSE.id,
             models.RSE.deleted == false())
    )

    if filters and filters.get('vo'):
        stmt = stmt.where(
            models.RSE.vo == filters.get('vo')
        )

    return [str(rseid) for rseid in session.execute(stmt).scalars().all()]



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import io
import mmap
import zlib
from functools import partial
from typing import TYPE_CHECKING

from rucio.common.bittorrent import merkle_sha256
from rucio.common.exception import ChecksumCalculationError

if TYPE_CHECKING:
    from _typeshed import FileDescriptorOrPath

# GLOBALLY_SUPPORTED_CHECKSUMS = ['adler32', 'md5', 'sha256', 'crc32']
GLOBALLY_SUPPORTED_CHECKSUMS = ['adler32', 'md5']
PREFERRED_CHECKSUM = GLOBALLY_SUPPORTED_CHECKSUMS[0]
CHECKSUM_KEY = 'supported_checksums'


def is_checksum_valid(checksum_name: str) -> bool:
    """
    A simple function to check whether a checksum algorithm is supported.
    Relies on GLOBALLY_SUPPORTED_CHECKSUMS to allow for expandability.

    :param checksum_name: The name of the checksum to be verified.
    :returns: True if checksum_name is in GLOBALLY_SUPPORTED_CHECKSUMS list, False otherwise.
    """

    return checksum_name in GLOBALLY_SUPPORTED_CHECKSUMS


def set_preferred_checksum(checksum_name: str) -> None:
    """
    If the input checksum name is valid,
    set it as PREFERRED_CHECKSUM.

    :param checksum_name: The name of the checksum to be verified.
    """
    if is_checksum_valid(checksum_name):
        global PREFERRED_CHECKSUM
        PREFERRED_CHECKSUM = checksum_name


def adler32(file: "FileDescriptorOrPath") -> str:
    """
    An Adler-32 checksum is obtained by calculating two 16-bit checksums A and B
    and concatenating their bits into a 32-bit integer. A is the sum of all bytes in the
    stream plus one, and B is the sum of the individual values of A from each step.

    :param file: file name
    :returns: Hexified string, padded to 8 values.
    """

    # adler starting value is _not_ 0
    adler = 1

    can_mmap = False
    # try:
    #    with open(file, 'r+b') as f:
    #        can_mmap = True
    # except:
    #    pass

    try:
        # use mmap if possible
        if can_mmap:
            with open(file, 'r+b') as f:
                m = mmap.mmap(f.fileno(), 0)
                # partial block reads at slightly increased buffer sizes
                for block in iter(partial(m.read, io.DEFAULT_BUFFER_SIZE * 8), b''):
                    adler = zlib.adler32(block, adler)
        else:
            with open(file, 'rb') as f:
                # partial block reads at slightly increased buffer sizes
                for block in iter(partial(f.read, io.DEFAULT_BUFFER_SIZE * 8), b''):
                    adler = zlib.adler32(block, adler)

    except Exception as e:
        raise ChecksumCalculationError('adler32', str(file), e)

    # backflip on 32bit -- can be removed once everything is fully migrated to 64bit
    if adler < 0:
        adler = adler + 2 ** 32

    return str('%08x' % adler)


def md5(file: "FileDescriptorOrPath") -> str:
    """
    Runs the MD5 algorithm (RFC-1321) on the binary content of the file named file and returns the hexadecimal digest

    :param file: file name
    :returns: string of 32 hexadecimal digits
    """
    hash_md5 = hashlib.md5()
    try:
        with open(file, "rb") as f:
            list(map(hash_md5.update, iter(lambda: f.read(4096), b"")))
    except Exception as e:
        raise ChecksumCalculationError('md5', str(file), e)

    return hash_md5.hexdigest()


def sha256(file: "FileDescriptorOrPath") -> str:
    """
    Runs the SHA256 algorithm on the binary content of the file named file and returns the hexadecimal digest

    :param file: file name
    :returns: string of 32 hexadecimal digits
    """
    with open(file, "rb") as f:
        bytes_ = f.read()  # read entire file as bytes
        readable_hash = hashlib.sha256(bytes_).hexdigest()
        print(readable_hash)
        return readable_hash


def crc32(file: "FileDescriptorOrPath") -> str:
    """
    Runs the CRC32 algorithm on the binary content of the file named file and returns the hexadecimal digest

    :param file: file name
    :returns: string of 32 hexadecimal digits
    """
    prev = 0
    for eachLine in open(file, "rb"):
        prev = zlib.crc32(eachLine, prev)
    return "%X" % (prev & 0xFFFFFFFF)


CHECKSUM_ALGO_DICT = {
    'adler32': adler32,
    'md5': md5,
    'sha256': sha256,
    'crc32': crc32,
    'merkle_sha256': merkle_sha256
}



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING, Optional, cast
from urllib.parse import urlparse

import qbittorrentapi

from rucio.common.config import get_rse_credentials
from rucio.common.constants import RseAttr
from rucio.common.utils import resolve_ip
from rucio.core.oidc import request_token
from rucio.db.sqla.constants import RequestState
from rucio.transfertool.transfertool import TransferStatusReport

from .bittorrent_driver import BittorrentDriver

if TYPE_CHECKING:
    from collections.abc import Sequence

    from sqlalchemy.orm import Session

    from rucio.common import types
    from rucio.core.rse import RseData


class QBittorrentTransferStatusReport(TransferStatusReport):

    supported_db_fields = [
        'state',
        'external_id',
    ]

    def __init__(self, request_id: str, external_id: str, qbittorrent_response: Optional[qbittorrentapi.TorrentDictionary]) -> None:
        super().__init__(request_id)

        if qbittorrent_response and qbittorrent_response.state_enum.is_complete == 1:
            new_state = RequestState.DONE
        else:
            new_state = RequestState.SUBMITTED

        self.state = new_state
        self.external_id = None
        if new_state in [RequestState.FAILED, RequestState.DONE]:
            self.external_id = external_id

    def initialize(self, session: "Session", logger: 'types.LoggerFunction' = logging.log) -> None:
        pass

    def get_monitor_msg_fields(self, session: "Session", logger: 'types.LoggerFunction' = logging.log) -> dict[str, str]:
        return {'protocol': 'qbittorrent'}


class QBittorrentDriver(BittorrentDriver):

    external_name = 'qbittorrent'
    required_rse_attrs = (RseAttr.QBITTORRENT_MANAGEMENT_ADDRESS, )

    @classmethod
    def make_driver(cls: "type[QBittorrentDriver]", rse: "RseData", logger: 'types.LoggerFunction' = logging.log) -> "Optional[BittorrentDriver]":

        address = rse.attributes.get(RseAttr.QBITTORRENT_MANAGEMENT_ADDRESS)
        if not address:
            return None

        url = urlparse(address)
        token = None
        if url.scheme.lower() == 'https':
            token = request_token(audience=url.hostname, scope='qbittorrent_admin')
        else:
            logging.debug(f'{cls.external_name} will not try token authentication. Requires HTTPS.')

        rse_cred = get_rse_credentials().get(rse.id, {})
        username = rse_cred.get('qbittorrent_username')
        password = rse_cred.get('qbittorrent_password')

        if not (token or (username and password)):
            return None

        return cls(
            address=address,
            username=username,
            password=password,
            token=token,
            logger=logger,
        )

    def __init__(self, address: str, username: str, password: str, token: Optional[str] = None, logger: 'types.LoggerFunction' = logging.log) -> None:
        extra_headers = None
        if token:
            extra_headers = {'Authorization': 'Bearer ' + token}

        self.client = qbittorrentapi.Client(
            host=address,
            username=username,
            password=password,
            EXTRA_HEADERS=extra_headers,
            FORCE_SCHEME_FROM_HOST=True,
        )
        self.logger = logger

    def listen_addr(self) -> tuple[str, int]:
        preferences = self.client.app_preferences()
        port = cast(int, preferences['listen_port'])
        ip = resolve_ip(urlparse(self.client.host).hostname or self.client.host)
        return ip, port

    def add_torrent(self, file_name: str, file_content: bytes, download_location: str, seed_mode: bool = False) -> None:
        self.client.torrents_add(
            rename=file_name,
            torrent_files=file_content,
            save_path=download_location,
            is_skip_checking=seed_mode,
            is_sequential_download=True,
        )

    def add_peers(self, torrent_id: str, peers: 'Sequence[tuple[str, int]]') -> None:
        self.client.torrents_add_peers(torrent_hashes=[torrent_id], peers=[f'{ip}:{port}' for ip, port in peers])

    def get_status(self, request_id: str, torrent_id: str) -> TransferStatusReport:
        info = self.client.torrents_info(torrent_hashes=[torrent_id])
        return QBittorrentTransferStatusReport(request_id, external_id=torrent_id, qbittorrent_response=info[0] if info else None)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
from typing import Optional


class Stopwatch:
    """Stopwatch to measure time durations.

    Note: The stopwatch is started on initialization.
    """

    _t_start: float
    _t_end: Optional[float]

    def __init__(self) -> None:
        self.restart()

    def _now(self) -> float:
        # TODO: change to time.monotonic_ns() if python 3.6 support is dropped.
        return time.monotonic()

    def restart(self) -> None:
        """Resets and starts the stopwatch."""
        self._t_start = self._now()
        self._t_end = None

    def stop(self) -> None:
        """Stops the stopwatch."""
        self._t_end = self._now()

    @property
    def elapsed(self) -> float:
        """Returns the total number of elapsed seconds."""
        if self._t_end is None:
            return self._now() - self._t_start
        else:
            return self._t_end - self._t_start

    def __float__(self) -> float:
        """Returns the total number of elapsed seconds."""
        return self.elapsed



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import logging
from os import path
from typing import TYPE_CHECKING, Any, Optional

from rucio.common.bittorrent import construct_torrent
from rucio.common.config import config_get
from rucio.common.constants import RseAttr
from rucio.common.extra import import_extras
from rucio.core.did_meta_plugins import get_metadata
from rucio.transfertool.transfertool import TransferStatusReport, Transfertool, TransferToolBuilder

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from rucio.common import types
    from rucio.core.request import DirectTransfer
    from rucio.core.rse import RseData

    from .bittorrent_driver import BittorrentDriver

DRIVER_NAME_RSE_ATTRIBUTE = 'bittorrent_driver'
DRIVER_CLASSES_BY_NAME: dict[str, type['BittorrentDriver']] = {}

EXTRA_MODULES = import_extras(['qbittorrentapi'])

if EXTRA_MODULES['qbittorrentapi']:
    from .bittorrent_driver_qbittorrent import QBittorrentDriver
    DRIVER_CLASSES_BY_NAME[QBittorrentDriver.external_name] = QBittorrentDriver


class BittorrentTransfertool(Transfertool):
    """
    Use bittorrent to perform the peer-to-peer transfer.
    """
    external_name = 'bittorrent'
    supported_schemes = {'magnet'}

    required_rse_attrs = (DRIVER_NAME_RSE_ATTRIBUTE, )

    def __init__(self, external_host: str, logger: 'types.LoggerFunction' = logging.log) -> None:
        super().__init__(external_host=external_host, logger=logger)

        self._drivers_by_rse_id = {}
        self.ca_cert, self.ca_key = None, None

        self.tracker = config_get('transfers', 'bittorrent_tracker_addr', raise_exception=False, default=None)

    @classmethod
    def _pick_management_api_driver_cls(cls: "type[BittorrentTransfertool]", rse: "RseData") -> Optional[type['BittorrentDriver']]:
        driver_cls = DRIVER_CLASSES_BY_NAME.get(rse.attributes.get(DRIVER_NAME_RSE_ATTRIBUTE, ''))
        if driver_cls is None:
            return None
        if not all(rse.attributes.get(attribute) is not None for attribute in driver_cls.required_rse_attrs):
            return None
        return driver_cls

    def _driver_for_rse(self, rse: "RseData") -> Optional['BittorrentDriver']:
        driver = self._drivers_by_rse_id.get(rse.id)
        if driver:
            return driver

        driver_cls = self._pick_management_api_driver_cls(rse)
        if not driver_cls:
            return None

        driver = driver_cls.make_driver(rse)
        self._drivers_by_rse_id[rse.id] = driver
        return driver

    @staticmethod
    def _get_torrent_meta(scope: "types.InternalScope", name: str) -> tuple[bytes, bytes, int]:
        meta = get_metadata(scope=scope, name=name, plugin='all')
        pieces_root = base64.b64decode(meta.get('bittorrent_pieces_root', ''))
        pieces_layers = base64.b64decode(meta.get('bittorrent_pieces_layers', ''))
        piece_length = meta.get('bittorrent_piece_length', 0)
        return pieces_root, pieces_layers, piece_length

    @classmethod
    def submission_builder_for_path(
            cls: "type[BittorrentTransfertool]",
            transfer_path: "list[DirectTransfer]",
            logger: 'types.LoggerFunction' = logging.log
    ) -> "tuple[list[DirectTransfer], Optional[TransferToolBuilder]]":
        hop = transfer_path[0]
        if hop.rws.byte_count == 0:
            logger(logging.INFO, f"Bittorrent cannot transfer fully empty torrents. Skipping {hop}")
            return [], None

        if not cls.can_perform_transfer(hop.src.rse, hop.dst.rse):
            logger(logging.INFO, f"The required RSE attributes are not set. Skipping {hop}")
            return [], None

        for rse in [hop.src.rse, hop.dst.rse]:
            driver_cls = cls._pick_management_api_driver_cls(rse)
            if not driver_cls:
                logger(logging.INFO, f"The rse '{rse}' is not configured correctly for bittorrent")
                return [], None

        pieces_root, _pieces_layers, piece_length = cls._get_torrent_meta(hop.rws.scope, hop.rws.name)
        if not pieces_root or not piece_length:
            logger(logging.INFO, "The required bittorrent metadata not set on the DID")
            return [], None

        return [hop], TransferToolBuilder(cls, external_host='Bittorrent Transfertool')

    def group_into_submit_jobs(self, transfer_paths: "Sequence[list[DirectTransfer]]") -> list[dict[str, Any]]:
        return [{'transfers': transfer_path, 'job_params': {}} for transfer_path in transfer_paths]

    @staticmethod
    def _connect_directly(torrent_id: str, peers_drivers: 'Sequence[BittorrentDriver]') -> None:
        peer_addr = []
        for i, driver in enumerate(peers_drivers):
            peer_addr.append(driver.listen_addr())

        for driver in peers_drivers:
            driver.add_peers(torrent_id=torrent_id, peers=peer_addr)

    def submit(self, transfers: "Sequence[DirectTransfer]", job_params: dict[str, str], timeout: Optional[int] = None) -> str:
        [transfer] = transfers
        rws = transfer.rws

        tracker = transfer.dst.rse.attributes.get(RseAttr.BITTORRENT_TRACKER_ADDR, self.tracker)

        src_drivers = {}
        for source in transfer.sources:
            driver = self._driver_for_rse(source.rse)
            if driver:
                src_drivers[source] = driver

        dst_driver = self._driver_for_rse(transfer.dst.rse)

        if not dst_driver or not src_drivers:
            raise Exception('Cannot initialize bittorrent drivers to submit transfers')

        pieces_root, pieces_layers, piece_length = self._get_torrent_meta(rws.scope, rws.name)
        torrent_id, torrent = construct_torrent(
            scope=str(rws.scope),
            name=rws.name,
            length=rws.byte_count,
            piece_length=piece_length,
            pieces_root=pieces_root,
            pieces_layers=pieces_layers,
            trackers=[tracker] if tracker else None,
        )

        for source, driver in src_drivers.items():
            source_protocol = transfer.source_protocol(source)
            [lfn] = source_protocol.parse_pfns([transfer.source_url(source)]).values()
            driver.add_torrent(
                file_name=rws.name,
                file_content=torrent,
                download_location=lfn['prefix'] + path.dirname(lfn['path']),
                seed_mode=True,
            )

        dest_protocol = transfer.dest_protocol()
        [lfn] = dest_protocol.parse_pfns([transfer.dest_url]).values()
        dst_driver.add_torrent(
            file_name=rws.name,
            file_content=torrent,
            download_location=lfn['prefix'] + lfn['path'],
        )

        self._connect_directly(torrent_id, [dst_driver] + list(src_drivers.values()))
        return torrent_id

    def bulk_query(self, requests_by_eid, timeout: Optional[int] = None) -> 'Mapping[str, Mapping[str, TransferStatusReport]]':
        response = {}
        for transfer_id, requests in requests_by_eid.items():
            for request_id, request in requests.items():
                driver = self._driver_for_rse(request['dst_rse'])
                if not driver:
                    self.logger(f'Cannot instantiate BitTorrent driver for {request["dest_rse"]}')
                    continue
                response.setdefault(transfer_id, {})[request_id] = driver.get_status(request_id=request_id, torrent_id=transfer_id)
        return response

    def query(self, transfer_ids: 'Sequence[str]', details: bool = False, timeout: Optional[int] = None) -> None:
        pass

    def cancel(self, transfer_ids: 'Sequence[str]', timeout: Optional[int] = None) -> None:
        pass

    def update_priority(self, transfer_id: str, priority: int, timeout: Optional[int] = None) -> None:
        pass



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' remove replicas_tombstone and replicas_rse_id indexes '''

from alembic.op import create_index, drop_index

# Alembic revision identifiers
revision = '27e3a68927fb'
down_revision = '295289b5a800'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    drop_index('REPLICAS_TOMBSTONE_IDX', 'replicas')
    drop_index('REPLICAS_RSE_ID_IDX', 'replicas')


def downgrade():
    '''
    Downgrade the database to the previous revision
    '''
    create_index('REPLICAS_RSE_ID_IDX', 'replicas', ['rse_id'])
    create_index('REPLICAS_TOMBSTONE_IDX', 'replicas', ['tombstone'])



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Graphite and prometheus metrics
"""

import atexit
import logging
import os
import string
from abc import abstractmethod
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union

from prometheus_client import REGISTRY, CollectorRegistry, Counter, Gauge, Histogram, generate_latest, multiprocess, push_to_gateway, start_http_server, values
from statsd import StatsClient

import __main__ as main
from rucio.common.config import config_get, config_get_bool, config_get_int
from rucio.common.stopwatch import Stopwatch
from rucio.common.utils import retrying

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence

    from rucio.common.types import LoggerFunction

_T = TypeVar('_T')
_M = TypeVar('_M', bound="_MultiMetric")

PROMETHEUS_MULTIPROC_DIR = os.environ.get('PROMETHEUS_MULTIPROC_DIR', os.environ.get('prometheus_multiproc_dir', None))


def cleanup_prometheus_files_at_exit() -> None:
    if PROMETHEUS_MULTIPROC_DIR:
        multiprocess.mark_process_dead(os.getpid())


class MultiprocessMutexValue(values.MultiProcessValue()):
    """
    MultiprocessValue protected by mutex

    Rucio usually is deployed using the apache MPM module, which means that it both uses multiple
    subprocesses, and multiple threads per subprocess.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = Lock()

    def inc(self, amount: float) -> None:
        with self._lock:
            return super().inc(amount)

    def set(self, value: float) -> None:
        with self._lock:
            return super().set(value)

    def get(self) -> float:
        with self._lock:
            return super().get()


if PROMETHEUS_MULTIPROC_DIR:
    os.makedirs(PROMETHEUS_MULTIPROC_DIR, exist_ok=True)
    values.ValueClass = MultiprocessMutexValue

    atexit.register(cleanup_prometheus_files_at_exit)


SERVER = config_get('monitor', 'carbon_server', raise_exception=False, default=None)
PORT = config_get_int('monitor', 'carbon_port', raise_exception=False, default=8125)
SCOPE = config_get('monitor', 'user_scope', raise_exception=False, default='rucio')
STATSD_CLIENT = None
if SERVER is not None:
    STATSD_CLIENT = StatsClient(host=SERVER, port=PORT, prefix=SCOPE)

ENABLE_METRICS = config_get_bool('monitor', 'enable_metrics', raise_exception=False, default=False)
if ENABLE_METRICS:
    METRICS_PORT = config_get_int('monitor', 'metrics_port', raise_exception=False, default=8080)
    start_http_server(METRICS_PORT, registry=REGISTRY)

COUNTERS = {}
GAUGES = {}
TIMINGS = {}
METRICS_LOCK = Lock()


_HISTOGRAM_DEFAULT_BUCKETS = Histogram.DEFAULT_BUCKETS


def _cleanup_old_prometheus_files(
        path: str,
        file_pattern: str,
        cleanup_delay: float,
        logger: "LoggerFunction"
) -> None:
    """cleanup behind processes which didn't finish gracefully."""

    oldest_accepted_mtime = datetime.now() - timedelta(seconds=cleanup_delay)
    for file in Path(path).glob(file_pattern):
        if not file.is_file():
            continue

        file_mtime = datetime.fromtimestamp(file.stat().st_mtime)

        if file_mtime < oldest_accepted_mtime:
            logger(logging.INFO, 'Cleaning up prometheus db file %s', file)
            try:
                os.remove(file)
            except FileNotFoundError:
                # Probably file already removed by another concurrent process
                pass


def cleanup_old_prometheus_files(logger: "LoggerFunction" = logging.log) -> None:
    path = PROMETHEUS_MULTIPROC_DIR
    if path:
        _cleanup_old_prometheus_files(path, file_pattern='gauge_live*.db', cleanup_delay=timedelta(hours=1).total_seconds(), logger=logger)
        _cleanup_old_prometheus_files(path, file_pattern='*.db', cleanup_delay=timedelta(days=7).total_seconds(), logger=logger)


@retrying(retry_on_exception=lambda _: True,
          wait_fixed=500,
          stop_max_attempt_number=2)
def generate_prometheus_metrics() -> bytes:
    cleanup_old_prometheus_files()

    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    return generate_latest(registry)


class _MultiMetric:
    """
    Thin wrapper class allowing to record both prometheus and statsd metrics.

    Inspired by the prometheus metric behavior: uses labels to parametrize metrics.
    In case of statsd, metrics are formatted using str.format(**labels). The prometheus
    ones using metric.labels(**labels) calls.

    If the prometheus metric string is not provided, it is derived from the statsd one.

    If `labelnames` is not provided, tries to extract them from the metric name by parsing
    it as a format string.
    """

    def __init__(
            self,
            statsd: str,
            prom: Optional[Union[str, Counter, Gauge, Histogram]] = None,
            documentation: Optional[str] = None,
            labelnames: Optional['Sequence[str]'] = None,
            registry: Optional[CollectorRegistry] = None
    ):
        """
        :param statsd: a string, eventually with keyword placeholders for the str.format(**labels) call
        :param prom: a string prometheus metric name; or an instantiated prometheus metric object
        """
        self._registry = registry or REGISTRY
        self._documentation = documentation or ''
        self._statsd = statsd
        if not prom:
            parsed_format = list(string.Formatter().parse(statsd))
            # automatically generate a prometheus metric name
            #
            # remove '.{label}' from the string for each `label`
            # substituted dots with underscores
            if labelnames is None:
                labelnames = tuple(field_name for _, field_name, _, _ in parsed_format if field_name)
            prom = ''.join(literal_text.rstrip('.').replace('.', '_') for literal_text, *_ in parsed_format)
        labelnames = labelnames or ()
        if isinstance(prom, str):
            self._prom = self.init_prometheus_metric(prom, self._documentation, labelnames=labelnames)
        else:
            self._prom = prom

        self._labelnames = labelnames

    @abstractmethod
    def init_prometheus_metric(self, name: str, documentation: Optional[str], labelnames: 'Sequence[str]' = ()):
        pass

    def labels(self: _M, **labelkwargs) -> _M:
        if not labelkwargs:
            return self

        return self.__class__(
            prom=self._prom.labels(**labelkwargs),  # type: ignore
            statsd=self._statsd.format(**labelkwargs),
            documentation=self._documentation,
            labelnames=self._labelnames,
            registry=self._registry,
        )


class _MultiCounter(_MultiMetric):

    def inc(self, delta: int = 1) -> None:
        delta = abs(delta)
        self._prom.inc(delta)  # type: ignore
        if STATSD_CLIENT:
            STATSD_CLIENT.incr(self._statsd, delta)

    def init_prometheus_metric(self, name: str, documentation: str, labelnames: 'Sequence[str]' = ()) -> Counter:
        return Counter(name, documentation, labelnames=labelnames, registry=self._registry)


class _MultiGauge(_MultiMetric):

    def set(self, value) -> None:
        self._prom.set(value)  # type: ignore
        if STATSD_CLIENT:
            STATSD_CLIENT.gauge(self._statsd, value)

    def init_prometheus_metric(self, name: str, documentation: str, labelnames: 'Sequence[str]' = ()) -> Gauge:
        return Gauge(name, documentation, labelnames=labelnames, registry=self._registry)


class _MultiTiming(_MultiMetric):

    def __init__(
            self,
            statsd: str,
            prom: Optional[str] = None,
            documentation: Optional[str] = None,
            labelnames: Optional['Sequence[str]'] = None,
            registry: Optional[CollectorRegistry] = None,
            buckets: 'Iterable[float]' = _HISTOGRAM_DEFAULT_BUCKETS,
    ) -> None:
        self._stopwatch = None
        self._histogram_buckets = tuple(buckets)
        super().__init__(statsd, prom, documentation, labelnames, registry)

    def observe(self, value: float) -> None:
        self._prom.observe(value)  # type: ignore
        if STATSD_CLIENT:
            STATSD_CLIENT.timing(self._statsd, value * 1000)

    def init_prometheus_metric(self, name: str, documentation: str, labelnames: 'Sequence[str]' = ()) -> Histogram:
        return Histogram(name, documentation, labelnames=labelnames, registry=self._registry, buckets=self._histogram_buckets)

    def __enter__(self):
        self._stopwatch = Stopwatch()
        return self

    def __exit__(self, typ, value, tb):
        if self._stopwatch:
            self._stopwatch.stop()
            self.observe(self._stopwatch.elapsed)


def _fetch_or_create_metric(
        name: str,
        labelnames: Optional['Sequence[str]'],
        container: dict[str, _T],
        factory: 'Callable[[str, Optional[Sequence[str]]], _T]'
) -> "_T":
    metric = container.get(name)
    if not metric:
        with METRICS_LOCK:
            metric = container.get(name)
            if not metric:
                container[name] = metric = factory(name, labelnames)
    return metric


def _fetch_or_create_counter(
        name: str,
        labelnames: Optional['Sequence[str]'] = None,
        documentation: Optional[str] = None,
        registry: Optional[CollectorRegistry] = None,
) -> _MultiCounter:
    return _fetch_or_create_metric(
        name=name,
        labelnames=labelnames,
        container=COUNTERS,
        factory=lambda _name, _labelnames: _MultiCounter(statsd=_name, labelnames=_labelnames,
                                                         documentation=documentation, registry=registry)
    )


def _fetch_or_create_gauge(
        name: str,
        labelnames: Optional['Sequence[str]'] = None,
        documentation: Optional[str] = None,
        registry: Optional[CollectorRegistry] = None,
) -> _MultiGauge:
    return _fetch_or_create_metric(
        name=name,
        labelnames=labelnames,
        container=GAUGES,
        factory=lambda _name, _labelnames: _MultiGauge(statsd=_name, labelnames=_labelnames,
                                                       documentation=documentation, registry=registry)
    )


def _fetch_or_create_timer(
        name: str,
        labelnames: Optional['Sequence[str]'] = None,
        documentation: Optional[str] = None,
        registry: Optional[CollectorRegistry] = None,
        buckets: 'Iterable[float]' = _HISTOGRAM_DEFAULT_BUCKETS
) -> _MultiTiming:
    return _fetch_or_create_metric(
        name=name,
        labelnames=labelnames,
        container=TIMINGS,
        factory=lambda _name, _labels: _MultiTiming(statsd=_name, labelnames=_labels, documentation=documentation,
                                                    registry=registry, buckets=buckets)
    )


class MetricManager:

    """
    Wrapper for metrics which prefixes them automatically with the given prefix or,
    alternatively, with the path of the module.
    """

    def __init__(self, prefix: Optional[str] = None, module: Optional[str] = None,
                 registry: Optional[CollectorRegistry] = None, push_gateways: Optional['Sequence[str]'] = None):
        if prefix:
            self.prefix = prefix
        elif module:
            self.prefix = module
        else:
            self.prefix = None
        self.registry = registry or REGISTRY
        self.push_gateways = push_gateways or []

    def full_name(self, name: str) -> str:
        if self.prefix:
            return f'{self.prefix}.{name}'
        return name

    def get_registry(self) -> CollectorRegistry:
        return self.registry

    def counter(
            self,
            name: str,
            *,
            labelnames: Optional['Sequence[str]'] = None,
            documentation: Optional[str] = None,
    ) -> _MultiCounter:
        """
        Log a counter.

        :param name: The name (suffix) of the counter to be retrieved
        :param labelnames: optional labels used to parametrize the metric
        :param documentation: optional prometheus documentation for this metric
        """
        return _fetch_or_create_counter(name=self.full_name(name), labelnames=labelnames, documentation=documentation)

    def gauge(
            self,
            name: str,
            *,
            labelnames: Optional['Sequence[str]'] = None,
            documentation: Optional[str] = None,
    ) -> _MultiGauge:
        """
        Log gauge information for a single stat

        :param name: The name (suffix) of the counter to be retrieved
        :param labelnames: optional labels used to parametrize the metric
        :param documentation: optional prometheus documentation for this metric
        """
        return _fetch_or_create_gauge(name=self.full_name(name), labelnames=labelnames, documentation=documentation)

    def timer(
            self,
            name: str,
            *,
            labelnames: Optional['Sequence[str]'] = None,
            documentation: Optional[str] = None,
            buckets: 'Iterable[float]' = _HISTOGRAM_DEFAULT_BUCKETS
    ) -> _MultiTiming:
        """
        Log a time measurement.

        :param name: The name (suffix) of the counter to be retrieved
        :param labelnames: optional labels used to parametrize the metric
        :param documentation: optional prometheus documentation for this metric
        :param buckets: Optional iterable of histogram bucket separators.
        """
        return _fetch_or_create_timer(name=self.full_name(name), labelnames=labelnames, documentation=documentation, buckets=buckets)

    def time_it(self, original_function=None, *, buckets=_HISTOGRAM_DEFAULT_BUCKETS):
        """
        Function decorator which records a timer: the amount of time spent in the function.
        """
        def _decorator(func):
            @wraps(func)
            def _wrapper(*args, **kwargs):
                with self.timer(name=func.__name__, buckets=buckets):
                    return func(*args, **kwargs)
            return _wrapper
        if original_function:
            return _decorator(original_function)
        return _decorator

    def count_it(self, original_function=None):
        """
        Function decorator which records a counter: how many times the function was executed.
        """
        def _decorator(func):
            @wraps(func)
            def _wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                finally:
                    _fetch_or_create_counter(name=self.full_name(func.__name__) + '_cnt').inc()
            return _wrapper
        if original_function:
            return _decorator(original_function)
        return _decorator

    def push_metrics_to_gw(self, job: Optional[str] = None, grouping_key: Optional[dict[str, Any]] = None) -> None:
        """
        Push the metrics out to the prometheus push gateways. This is useful for short-running programs which don't
        live long enough to be reliably scraped in the prometheus pull model.
        """

        if not job:
            job = Path(main.__file__).stem
        grouping_key = grouping_key or {}

        for server in self.push_gateways:
            try:
                push_to_gateway(server.strip(), job=job, registry=self.registry, grouping_key=grouping_key)
            except:
                continue
